
from pycamia import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "micomputing",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File to visualize features of 3D volume.",
    help = "Use `from micomputing import plot as plt`.",
    requires = "matplotlib"
).check()

__all__ = """
    set_font
    plot
    subplots
    background
    imshow
    imsshow
    volshow
    maskshow
    bordershow
    fieldshow
    gridshow
    
    TensorBoard tb
""".split()

import math
from .trans import DenseDisplacementField, Transformation
from .stdio import IMG
from .funcs import reorient

with __info__:
    import batorch as bt
    import matplotlib as mpl
    from matplotlib import pyplot as plt
    from matplotlib import colors as mc
    from matplotlib.pyplot import *
    from pyoverload import *
    from pycamia import prod, to_tuple, argmax, avouch, touch, tokenize
    __all__.extend([x for x in dir(plt) if not x.startswith('_') and x not in __all__])
    mpl.rcParams['agg.path.chunksize'] = 0
    plt.rcParams['font.sans-serif'] = 'Arial Unicode MS'
    plt.rcParams['axes.unicode_minus'] = False
    plt.rcParams['savefig.dpi'] = 300

canvas = None
colors = ['crimson', 'forestgreen', 'royalblue', 'red', 'green', 'blue', 'gold', 'purple', 'gray', 'pink', 'darkgreen', 'dodgerblue']
assert all(c in mc.CSS4_COLORS for c in colors)
subplot_rows = 1
subplot_cols = 1
subplot_id = 1
default_orient = None

TRS_list = ["trans", "trs", "transversal", "transverse", "ax", "axial", 
            "axis", "crs", "cross", "hor", "horizontal", "slice"]
COR_list = ["cor", "coro", "coronal", "coron"]
SAG_list = ["sag", "sagit", "sagittal", "vert", "vertical"]
name_to_orient = {k: 'AR' for k in TRS_list}
name_to_orient.update({k: 'SR' for k in COR_list})
name_to_orient.update({k: 'SA' for k in SAG_list})

def set_font(font: str):
    plt.rcParams['font.sans-serif'] = font
    
def set_orient(ref):
    if isinstance(ref, IMG): ref = ref.orientation
    global default_orient
    default_orient = ref

def to_image(data: array, nslice: (int, null)=None, dim: int=-1, as_orient=None, to_orient='AR'):
    if as_orient is None: as_orient = default_orient
    if to_orient is None: to_orient = 'AR' # default orient for output
    as_orient = name_to_orient.get(as_orient.lower(), as_orient.upper()) if as_orient else None
    to_orient = name_to_orient.get(to_orient.lower(), to_orient.upper()) if to_orient else None
    axis_upper = {'L':'R', 'I':'S', 'P':'A'}
    to_orilist = [axis_upper.get(o, o) for o in to_orient]
    if as_orient and len(as_orient) == 3 and len(to_orilist) < 3:
        to_orilist += [x for x in "ARS" if x not in to_orilist]
        to_orient += to_orilist[-1]
    if isinstance(data, IMG):
        if as_orient is None:
            data = data.reorient(to_orient)
            if data.n_dim == 3 and len(to_orilist) < 3:
                to_orilist += [x for x in "ARS" if x not in to_orilist]
                to_orient += to_orilist[-1]
            if isinstance(dim, int): # LPI => to_orient
                from_orilist = [axis_upper.get(o, o) for o in data.orientation]
                remapper = {k:to_orilist.index(from_orilist[k]) for k in range(-3, 3)}
                dim = remapper[dim]
        data = data.to_tensor()
    sp = getattr(data, 'spacing', None)
    if not isinstance(data, bt.torch.Tensor): data = bt.tensor(data)
    elif not isinstance(data, bt.Tensor): data = data.as_subclass(bt.Tensor).init_special()
    data = data.squeeze()
    if sp is not None:
        for d, s in enumerate(sp):
            data = data.amplify(round(s), d)
    if as_orient is not None: data = reorient(data, from_orient=as_orient, to_orient=to_orient)
    if isinstance(dim, str) and len(dim) == 2: dim = argmax([o in dim for o in to_orient])
    if data.n_dim <= 1: raise TypeError("Please don't use 'plot.imshow' to demonstrate an array or a scalar. ")
    if data.n_space_dim > 3: raise TypeError(f"'plot.imshow' takes 2 or 3D-data as input (currently {data.shape}), please reduce the dimension manually or specify special dimensions to reduce. ")
    if data.n_space_dim == 3:
        if data.has_batch: data = data.pick(0, {})
        if data.has_channel: 
            if dim < 0: dim += data.n_space_dim
            dim = data.space_start + dim
            if nslice is None: nslice = data.size(dim) // 2
            data = data.pick(nslice, dim) \
                       .sample(number=min(data.channel_size, 3), random=False, dim=[]) \
                       .movedim([], -1)
        else:
            if data.space[-1] <= 3: pass
            elif data.space[0] <= 3: data = data.movedim(0, 2)
            if nslice is None: nslice = data.size(dim) // 2
            if dim < 0: dim += data.n_space_dim
            dim = data.space_start + dim
            data = data.pick(nslice, dim)
    elif data.n_space_dim == 2:
        if data.has_batch: data = data.pick(0, {})
        if data.has_channel:
            data = data.sample(number=min(data.channel_size, 3), random=False, dim=[]) \
                       .movedim(data.channel_dimension, -1)
    elif data.n_dim == 3: data = data.pick(0, {}).init_special()
    elif data.n_dim == 2: data = data.init_special()
    else: TypeError(f"Cannot extrat image from data of size {data.shape}. ")
    if data.n_dim == 3 and data.size(2) < 3: data = bt.cat(data, bt.zeros(data.shape[:-1] + (3 - data.size(2),)).special_from(data), 2)
    if data.n_dim == 2 and data.max() == data.min(): data = data.duplicated(3, -1)
    return data.float()

def to_RGB(*color):
    if len(color) == 0: return (1.,) * 3
    elif len(color) == 1:
        c = color[0]
        if isinstance(c, float) and 0 <= c <= 1: return (c,) * 3
        elif isinstance(c, (int, float)) and 0 <= c <= 255: return (c / 255,) * 3
        elif isinstance(c, tuple): return to_RGB(*c)
        elif isinstance(c, str):
            if not c.startswith('#'):
                c = mc.BASE_COLORS.get(c.lower(), mc.CSS4_COLORS.get(c.lower(), c))
                if isinstance(c, tuple): return to_RGB(*c)
            if not c.startswith('#'): raise TypeError(f"Unrecognized color {c}.")
            return mc.hex2color(c)
        else: raise TypeError("Unaccepted color type. ")
    elif len(color) == 3:
        if all(isinstance(c, float) and 0 <= c <= 1 for c in color): return color
        elif all(isinstance(c, (int, float)) and 0 <= c <= 255 for c in color): return tuple(c / 255 for c in color)
        else: raise TypeError("Unaccepted color type. ")
    else: raise TypeError("Unaccepted color type. ")

def subplots(n=1, m=None):
    if m is None:
        r = math.floor(math.sqrt(n))
        c = math.ceil(n / r)
        n, m = r, c
    global subplot_rows
    global subplot_cols
    global subplot_id
    subplot_rows = n
    subplot_cols = m
    subplot_id = 1

def clf():
    plt.clf()
    subplots()
    
def create_subplot():
    global subplot_rows
    global subplot_cols
    global subplot_id
    if subplot_id > subplot_rows * subplot_cols:
        raise TypeError(f"Cannot create subplot {subplot_id} in grid {subplot_rows} x {subplot_cols}, " + 
                        "please use `micomputing.plot.subplots(rows, cols)` to specify the number of images. ")
    plt.subplot(subplot_rows, subplot_cols, subplot_id); subplot_id += 1

def plot(*data: (array, null), **kwargs):
    """
    An alias of matplotlib.pyplot.plot, compatible with batorch. 
    
    Args: data (bt.Tensor): {n_batch} for multiple lines, while [n_channel] for x and y.
    """
    create_subplot()
    return line(*data, **kwargs)

def line(*data: (array, null), **kwargs):
    """
    Args: data (bt.Tensor): {n_batch} for multiple lines, while [n_channel] for x and y.
    """
    if len(data) == 1: 
        y = bt.tensor(data[0])
        avouch(y.n_space_dim == 1)
        if y.n_channel == 2: x, y = [u.squeeze() for u in y.split(1, [])]
        else:
            length = y.space[0]
            x = bt.arange(0, length).expand_to(y)
    elif len(data) == 2:
        x, y = data
    else: raise TypeError("Too much arguments for micomputing.plot/line. ")
    if isinstance(x, bt.torch.Tensor):
        if x.device.type == 'cuda': x = x.cpu()
        x = x.detach().cpu().numpy()
    if isinstance(y, bt.torch.Tensor):
        if y.device.type == 'cuda': y = y.cpu()
        y = y.detach().cpu().numpy()
    return plt.plot(x.T, y.T, **kwargs)

@typehint
def imshow(data: (array, null)=None, nslice: (int, null)=None, dim: int=-1, title=None, **kwargs):
    """
    An automatic image display function for all kinds of tensors. 
    The first image in batched images will be selected to be showed. 
    For medical images:
    Displacements with channel dimension identified will be displayed as RGB colored maps.
    If there are no dimension <=3, gray scaled images will be showed. 
    Transverse medical image with the right hand side of the subject shown on the left
        and anterior shown at the bottom will be selected for 3D volumes.
    `nslice` and `dim` are used for 3D volumes only, meaning to show the `nslice` slice of dimension `dim`. 
    """
    global canvas
    kwargs = kwargs.copy()
    if kwargs.pop('create_subplot', data is not None): create_subplot()
    if data is not None:
        canvas = to_image(data, nslice, dim, kwargs.pop('as_orient', None), kwargs.pop('to_orient', 'ARS'))
    if canvas is None or isinstance(canvas, tuple):
        raise TypeError("Please input data in 'imshow' or 'background' to show. ")
    if isinstance(canvas, bt.torch.Tensor):
        if canvas.device.type == 'cuda': canvas_show = canvas.cpu()
        canvas_show = canvas.detach().cpu().numpy()
    else: canvas_show = canvas
    plt.axis('off')
    kwargs['cmap'] = kwargs.get('cmap', 'gray')
    if canvas_show.ndim >= 3:
        Max = canvas_show.max()
        if canvas_show.dtype.kind == 'i' and Max > 255:
            canvas_show = canvas_show * 255 // canvas_show.max()
            if title is None: title = f"x {Max / 255:.04f}"
        elif Max > 1:
            canvas_show = canvas_show / canvas_show.max()
            if title is None: title = f"x {Max:.04f}"
        if title is None: title = ''
    plt.title(title)
    return plt.imshow(canvas_show, **kwargs)

@typehint
def imsshow(*data: (array, null), nslice: (int, null)=None, dim: int=-1, show=False, **kwargs):
    """
    Display a series of images by 'imshow'. Use argument 'show=True' to directly call plt.show(). 
    The first image in batched images will be selected to be showed. 
    For medical images:
    Displacements with channel dimension identified will be displayed as RGB colored maps.
    If there are no dimension <=3, gray scaled images will be showed. 
    Transverse medical image with the right hand side of the subject shown on the left
        and anterior shown at the bottom will be selected for 3D volumes.
    `nslice` and `dim` are used for 3D volumes only, meaning to show the `nslice` slice of dimension `dim`. 
    """
    return [imshow(d, nslice, dim, **kwargs) for d in data] + ([plt.show()] if show else [])

@typehint
def volshow(data: (array, null), show=False, **kwargs):
    subplots(2, 2)
    return [imshow(data, None, (d + 2) % 3, **kwargs) for d in range(len(data.shape))]

def background(*color):
    """
    Set a background color by RGB or a gray scale, conflict with imshow. 
    """
    global canvas
    canvas = to_RGB(*color)
    create_subplot()

@typehint
def maskshow(*masks, on=None, alpha=0.5, nslice=None, dim=-1, stretch=False, **kwargs):
    global canvas
    orient_as = kwargs.pop('as_orient', None)
    orient_to = kwargs.pop('to_orient', None)
    if on is not None:
        if isinstance(on, (int, tuple)): background(*on)
        elif isinstance(on, array): canvas = to_image(on, nslice, dim, orient_as, orient_to)
        elif isinstance(on, list): canvas = to_image(bt.tensor(on), nslice, dim, orient_as, orient_to)
        else: raise TypeError("Unrecognized argument 'on' for 'maskshow'. ")
    elif canvas is None:
        canvas = (1.,) * 3
    if len(masks) == 1 and not isinstance(masks[0], array):
        raise TypeError("Non-array mask for function 'maskshow', unzip the tuple if necessary. ")
    if len(masks) == 0: return imshow(create_subplot = on is not None)
    alpha = to_tuple(alpha, len(masks))
    new_masks = []
    new_alpha = []
    for m, a in zip(masks, alpha):
        img = to_image(m, nslice, dim, orient_as, orient_to)
        if img.n_dim == 3:
            new_masks.extend(x.squeeze(-1) for x in img.split(1, dim=dim))
            new_alpha.extend([a] * img.size(dim))
        else:
            new_masks.append(img)
            new_alpha.append(a)
    color_mask_map = [(to_RGB(c), m, a) for c, m, a in zip(colors*(len(new_masks) // len(colors) + 1), new_masks, new_alpha)]
    to_pop = []
    for c, m in kwargs.items():
        if touch(lambda: to_RGB(c)) is not None: 
            color_mask_map.append((to_RGB(c), m, alpha[0]))
            to_pop.append(c)
    for c in to_pop: kwargs.pop(c)

    if not stretch:
        shapes = [m.size() for _, m, _ in color_mask_map]
        target_shape = shapes[0]
        if len(set(shapes)) > 1 or not isinstance(canvas, tuple) and target_shape != canvas.shape:
            raise TypeError(f"Please use masks (size: {target_shape}) of the same size as the background image (size: {canvas.shape}), "
                            "or use 'stretch=True' in 'maskshow' to automatically adjust the image sizes. ")
    else:
        def adjust(m, to):
            ms = tuple(m.shape)
            scaling = tuple((a // b, b // a) for a, b in zip(to, ms))
            return m.down_scale([max(v, 1) for u, v in scaling]).up_scale([max(u, 1) for u, v in scaling]).crop_as(to)
        shapes = [m.size() for _, m, _ in color_mask_map]
        if not isinstance(canvas, tuple): shapes.append(canvas.shape[:2])
        areas = [u * v for u, v in shapes]
        target_shape = shapes[areas.index(max(areas))]
        color_mask_map = [(c, adjust(m, to=target_shape), a) for c, m, a in color_mask_map]
        canvas = adjust(canvas, to=target_shape)

    target_shape = bt.Size(*target_shape, [3])
    if isinstance(canvas, tuple): canvas = bt.channel_tensor(canvas).expand_to(target_shape)
    elif canvas.n_dim == 2: canvas = canvas.expand_to(target_shape)
    coeff = prod(1 - a * m for _, m, a in color_mask_map) # color_mask_map: (color, mask, alpha)
    canvas *= coeff
    for i, (c, m, a) in enumerate(color_mask_map):
        coeff = prod(a * m if j == i else 1 - a * m for j, (_, m, a) in enumerate(color_mask_map))
        canvas += (coeff * m).unsqueeze([-1]) * bt.channel_tensor(c)
    if isinstance(canvas, bt.torch.Tensor):
        if canvas.device.type == 'cuda': canvas_show = canvas.cpu()
        canvas_show = canvas.detach().cpu().numpy()
    else: canvas_show = canvas
    plt.axis('off')
    create_subplot()
    kwargs['cmap'] = kwargs.get('cmap', 'gray')
    if canvas_show.ndim >= 3:
        Max = canvas_show.max()
        if canvas_show.dtype.kind == 'i' and Max > 255: canvas_show = canvas_show * 255 // canvas_show.max(); plt.title(f"x {Max / 255:.04f}")
        elif Max > 1: canvas_show = canvas_show / canvas_show.max(); plt.title(f"x {Max:.04f}")
    return plt.imshow(canvas_show, **kwargs)

def smooth(curve):
    """
    curve: 2xn
    """
    middle = (curve[:, :-2] + curve[:, 1:-1] + curve[:, 2:]) / 3
    if all(curve[:, 0] == curve[:, -1]):
        head = (curve[:, :1] + curve[:, 1:2] + curve[:, -2:-1]) / 3
        return bt.cat(head, middle, head, dim=1)
    return bt.cat(curve[:, :1], middle, curve[:, -1:], dim=1)

def sharpen(curve, old_curve):
    """
    sharpen towards old_curve
    curve, old_curve: 2xn
    """
    a = bt.cat(curve[:, -2:-1] - curve[:, -1:], curve[:, :-1] - curve[:, 1:], dim=1)
    b = bt.cat(curve[:, 1:] - curve[:, :-1], curve[:, 1:2] - curve[:, :1], dim=1)
    costheta = (a*b).sum(0) / bt.sqrt((a*a).sum(0)) / bt.sqrt((b*b).sum(0))
    new_curve = curve.clone()
    new_curve[:, costheta > -0.9] = old_curve[:, costheta > -0.9]
    return new_curve

def constraint(new_curve, old_curve, constraint_curve):
    """
    constraint new_curve towards constraint_curve
    curve, old_curve: 2xn
    """
    dis_sqs = bt.sum((new_curve - constraint_curve) ** 2, 0)
    percentile = 2 #min(np.sort(dis_sqs)[98 * len(dis_sqs) // 100], 1)
    return bt.where(dis_sqs <= percentile, new_curve, old_curve)

def border(mask, min_length = 10):
    grid = bt.image_grid(*mask.shape)
    mask = mask > 0.5
    idx = mask[1:, :] ^ mask[:-1, :]
    idx = idx.expand_to(2, -1, mask.size(1))
    locs1 = (grid[:, 1:, :] + grid[:, :-1, :])[idx] / 2
    idx = mask[:, 1:] ^ mask[:, :-1]
    idx = idx.expand_to(2, mask.size(0), -1)
    locs2 = (grid[:, :, 1:] + grid[:, :, :-1])[idx] / 2
    locs = bt.cat(locs1.reshape(2, -1), locs2.reshape(2, -1), dim=1)
    if locs.size == 0: return []
    curves = []
    unvisited = bt.ones(locs.shape[-1])
    while True:
        if not any(unvisited): break
        first = bt.argmax(unvisited).item()
        cloc = locs[:, first:first + 1]
        unvisited[first] = 0
        curve = cloc
        while True:
            dissq = bt.sum((locs - cloc) ** 2, 0)
            inloc = bt.argmax(bt.where((unvisited > 0) & (dissq > 0), 1/dissq.clamp(min=0.1), bt.tensor(0).float()))
            if dissq[inloc] > 2: break
            cloc = locs[:, inloc:inloc + 1]
            curve = bt.cat(curve, cloc, dim=1)
            unvisited[inloc] = 0
            if not any(unvisited): break
        sloc = locs[:, first:first + 1]
        if bt.sum((cloc - sloc) ** 2) <= 2:
            curve = bt.cat(curve, sloc, dim=1)
        if curve.shape[1] <= min_length: continue
        scurve = curve
        for _ in range(100): scurve = constraint(smooth(scurve), scurve, curve)
        ccurve = scurve
        for _ in range(100):
            scurve = constraint(sharpen(scurve, curve), scurve, ccurve)
            scurve = constraint(smooth(scurve), scurve, curve)
        curves.append(scurve)
    return curves

def bordershow(*masks, on=None, mask_alpha=0., nslice=None, dim=-1, stretch=False, min_length = 10, **kwargs):
    global canvas
    orient_as = kwargs.pop('as_orient', None)
    orient_to = kwargs.pop('to_orient', None)
    if on is not None:
        if isinstance(on, (int, tuple)): background(*on)
        elif isinstance(on, array): canvas = to_image(on, nslice, dim, orient_as, orient_to)
        elif isinstance(on, list): canvas = to_image(bt.tensor(on), nslice, dim, orient_as, orient_to)
        else: raise TypeError("Unrecognized argument 'on' for 'maskshow'. ")
    elif canvas is None:
        canvas = (1.,) * 3
    if len(masks) == 0: return
    new_masks = []
    for m in masks:
        img = to_image(m, nslice, dim, orient_as, orient_to)
        if img.n_dim == 3:
            new_masks.extend(x.squeeze(-1) for x in img.split(1, dim=dim))
        else:
            new_masks.append(img)
    color_mask_map = [(to_RGB(c), m) for c, m in zip(colors*(len(new_masks) // len(colors) + 1), new_masks)]
    for c, m in kwargs.items():
        if touch(lambda: to_RGB(c)) is not None: 
            color_mask_map.append((to_RGB(c), m))
            kwargs.pop(c)

    new_masks = [m for _, m in color_mask_map]
    shapes = [m.size() for _, m in color_mask_map]
    if not stretch:
        target_shape = shapes[0]
        if len(set(shapes)) > 1 or not isinstance(canvas, tuple) and target_shape != canvas.shape[:2]:
            raise TypeError("Please use masks of the same size as the background image, "
                            "or use 'stretch=True' in 'maskshow' to automatically adjust the image sizes. ")
    else:
        def adjust(m, to):
            ms = tuple(m.shape)
            scaling = tuple((a // b, b // a) for a, b in zip(to, ms))
            return m.down_scale([max(v, 1) for u, v in scaling]).up_scale([max(u, 1) for u, v in scaling]).crop_as(to)
        if not isinstance(canvas, tuple): shapes.append(canvas.shape[:2])
        areas = [u * v for u, v in shapes]
        target_shape = shapes[areas.index(max(areas))]
        color_mask_map = [(c, adjust(m, to=target_shape)) for c, m in color_mask_map]
        canvas = adjust(canvas, to=target_shape)
    if isinstance(canvas, tuple):
        canvas = bt.channel_tensor(canvas).expand_to(target_shape + bt.Size([3],)).with_channeldim(None)

    plots = []
    if on is not None: plots.append(imshow(create_subplot=True))
    if mask_alpha > 0: plots.append(maskshow(*new_masks, alpha=mask_alpha))
    else: plots.append(imshow())
    for color, mask in color_mask_map:
        curves = border(mask, min_length)
        for c in curves:
            if isinstance(c, bt.torch.Tensor):
                if c.device.type == 'cuda': c = c.cpu()
                c = c.detach().cpu().numpy()
            plots.append(plt.plot(c[1], c[0], color = color, **kwargs))
    return plots

@bt.batorch_wrapper
def fieldshow(disp: bt.Tensor, on: bt.Tensor=None, nslice=None, dim=-1, stretch=False, **kwargs):
    """
    fieldshow shows displacements with RGB channels to form a facke colored map 
        indicating different directions by different color and magnitude by brightness.
    """
    global canvas
    orient_as = kwargs.pop('as_orient', None)
    orient_to = kwargs.pop('to_orient', None)
    if on is not None:
        if isinstance(on, (int, tuple)): background(*on)
        elif isinstance(on, array): canvas = to_image(on, nslice, dim, orient_as, orient_to)
        elif isinstance(on, list): canvas = to_image(bt.tensor(on), nslice, dim, orient_as, orient_to)
        else: raise TypeError("Unrecognized argument 'on' for 'maskshow'. ")
    elif canvas is None:
        canvas = (1.,) * 3
	
    if isinstance(disp, Transformation):
        if canvas is None or isinstance(canvas, tuple) or not hasattr(on, 'n_dim') or on.n_dim != disp.n_dim:
            raise TypeError("Please input data in 'imshow' or 'background' before trying to display grid. ")
        disp = disp.toDDF(on.space).pick(0, {})
    elif canvas is None or isinstance(canvas, tuple): canvas = bt.zeros(disp.space)
    disp = disp.abs()
    if orient_as is not None: disp = reorient(disp, from_orient=orient_as, to_orient=orient_to)
    disp = to_image(disp, nslice, dim)
    if disp.size(2) > 3: disp = disp[..., :3]
    return imshow(disp, create_subplot = on is not None, **kwargs)

@bt.batorch_wrapper
def gridshow(disp, on: bt.Tensor=None, gap=None, nslice=None, dim=-1, stretch=False, origin=None, title=None, **kwargs):
    """
    gridshow shows grid for the displacements. 
    """
    global canvas
    orient_as = kwargs.pop('as_orient', None)
    orient_to = kwargs.pop('to_orient', 'ARS')
    if on is not None:
        if isinstance(on, (int, tuple)): background(*on)
        elif isinstance(on, array): canvas = to_image(on, nslice, dim, orient_as, orient_to)
        elif isinstance(on, list): canvas = to_image(bt.tensor(on), nslice, dim, orient_as, orient_to)
        else: raise TypeError("Unrecognized argument 'on' for 'maskshow'. ")
    elif canvas is None:
        canvas = (1.,) * 3
	
    if isinstance(disp, Transformation):
        if canvas is None or isinstance(canvas, tuple) or not hasattr(on, 'n_dim') or on.n_space_dim != disp.n_dim:
            raise TypeError("Please input data for displacement in 'imshow' or 'background' before trying to display grid: need size of the canvas. ")
        disp = disp.to_DDF(on.space)
    elif canvas is None or isinstance(canvas, tuple): canvas = bt.zeros(disp.space)
    disp = bt.to_bttensor(disp)
    if disp.has_batch: disp = disp.sample(random=False, dim={})
    if orient_as is not None: disp = reorient(disp, from_orient=orient_as, to_orient=orient_to)
    Y = DenseDisplacementField(disp)(bt.image_grid(disp.space)).squeeze({})
    Y = to_image(Y, nslice, dim).movedim(-1, 0)
    if Y.size(0) > 2:
        if dim < 0: dim += Y.size(0)
        Y = bt.cat(Y[:dim], Y[dim+1:], [])
        
    if origin is None: origin = 0
    if not isinstance(origin, (tuple, list)): origin = (origin,)
    if len(origin) == 1: origin = tuple(origin) * Y.size(0)
    Y = bt.channel_tensor(origin) + Y

    plots = [imshow(canvas, create_subplot = on is not None, title=title)]
    if on is None and (canvas is None or isinstance(canvas, (int, tuple))):
        plt.ylim((Y.space[0], 0))
        plt.xlim((0, Y.space[1]))
    if gap is None: gap = Y.space[0] // 8, Y.space[1] // 8
    gap = to_tuple(gap)
    if len(gap) == 1: gap = gap * 2
    if 'color' not in kwargs: kwargs['color'] = 'gold'
    if isinstance(Y, bt.torch.Tensor):
        if Y.device.type == 'cuda': Y = Y.cpu()
        Y_data = Y.detach().cpu().numpy()
    for row in list(bt.arange(0, Y.space[0], gap[0])) + [Y.space[0] - 1]:
        plots.append(plt.plot(Y_data[1, row, :], Y_data[0, row, :], **kwargs))
    for col in list(bt.arange(0, Y.space[1], gap[0])) + [Y.space[1] - 1]:
        plots.append(plt.plot(Y_data[1, :, col], Y_data[0, :, col], **kwargs))
    return plots

from torch.utils.tensorboard import SummaryWriter
from pycamia import Path, get_args_expression
import os, threading, subprocess

class TensorBoard:
    
    def __init__(self, log_dir=None, port=6006):
        if log_dir is None: log_dir = Path.curdir / 'logs'
        # self.command = f"tensorboard --logdir=./{log_dir - Path.curdir} --port={port}"
        self.writer = SummaryWriter(log_dir)
        # self.process = subprocess.Popen(self.command.split(), cwd=Path.curdir)
        self.global_step = {}
        
    def __del__(self):
        self.writer.close()
        # self.process.kill()
        
    def __enter__(self):
        return self
        
    def __exit__(self, error_type, error_msg, traceback): ...

    def add_image(self, *data, **kwargs):
        name = tokenize(get_args_expression(), ',')[0]
        gs = kwargs.pop('global_step', self.global_step.get(f'image-{name}', 0))
        img_kwargs = {k: kwargs.pop(k) for k in ['nslice', 'dim', 'as_orient', 'to_orient'] if k in kwargs}
        image_tensors = []
        for dt in data:
            dt = to_image(dt, **img_kwargs)
            if dt.min().item() >= -0.1 and dt.max().item() <= 1.1: dt = dt.clamp(0, 1)
            else: dt = bt.normalize01(dt)
            image_tensors.append(dt)
        image_tensor = bt.cat(image_tensors, -1)
        if image_tensor.has_feature: self.writer.add_image(kwargs.pop('tag', 'untitled'), image_tensor.move_dim([], 0), gs)
        else: self.writer.add_image(kwargs.pop('tag', 'untitled'), image_tensor.unsqueeze([0]), gs)
        self.global_step[f'image-{name}'] = gs + 1
        
    def add_scalar(self, value, **kwargs):
        name = tokenize(get_args_expression(), ',')[0]
        gs = kwargs.pop('global_step', self.global_step.get(f'scalar-{name}', 0))
        self.writer.add_scalar(kwargs.pop('tag', 'untitled'), value, gs)
        self.global_step[f'scalar-{name}'] = gs + 1
        
    def add_figure(self, value, **kwargs):
        name = tokenize(get_args_expression(), ',')[0]
        gs = kwargs.pop('global_step', self.global_step.get(f'figure-{name}', 0))
        self.writer.add_figure(kwargs.pop('tag', 'untitled'), value, gs)
        self.global_step[f'figure-{name}'] = gs + 1

# try: import tensorflow as tf; tf_available = True
# except ModuleNotFoundError: tf_available = False

# if tf_available:
#     class TensorBoard:
#         def __init__(self, env='vis-main', port=8097, **kwargs):
#     tb = TensorBoard()

# try: import visdom; visdom_available = True
# except ModuleNotFoundError: visdom_available = False

# if visdom_available:
#     import os
#     import subprocess
#     from pycamia import get_args_expression
    
#     class Visualizer:
#         def __init__(self, env='vis-main', port=8097, **kwargs):
#             "python3 -m visdom.server -port 8097"
#             # self.process = subprocess.Popen(['python3', '-m', 'visdom.server', '-port', str(port)])
#             self.viz = visdom.Visdom(env=env, port=port, **kwargs)

#             self.index = {}
#             self.log_text = ''
            
#         # def __del__(self): self.process.kill()

#         def img(self, data, **kwargs):
#             name = tokenize(get_args_expression(), ',')[0]
#             img_kwargs = {k: kwargs.pop(k) for k in ['nslice', 'dim', 'as_orient', 'to_orient'] if k in kwargs}
#             if data.min().item() >= -0.1 and data.max().item() <= 1.1: data = data.clamp(0, 1)
#             else: data = bt.normalize01(data)
#             self.viz.image(to_image(data, **img_kwargs), win=name, opts=dict(title=name, store_history=True), **kwargs)

#         def text(self, text, name='text'):
#             self.viz.text(text, win=name)
        
#         def imgs(self, images, win_name='train', nrow=2, img_name=None):
#             """
#             There are images in on win.
#             images: single image or concated images
#             win_name: the window name
#             nrow: number of images in a row
#             img_name: window title name

#             Example:
#             # only show one image in a batch
#             images = torch.cat([input[0], gt[0], output[0])
#             self.vis.images(images, win_name='train', img_name=img_name[0], nrow=3)
#             """
#             if img_name is None:
#                 title = win_name
#             else:
#                 title = '{}_{}'.format(win_name, img_name)
#             self.viz.images(
#                 images,
#                 nrow=nrow,
#                 win=win_name,
#                 opts=dict(title=title, caption=img_name)
#             )

#         def __getattr__(self, name):
#             '''
#             self.function self.vis.function
#             plot,image,log,plot_many
#             '''
#             return getattr(self.vis, name)

#         # plot line
#         def plot_multi_win(self, d, loop_flag=None):
#             '''

#             @params d: dict (name, value) i.e. ('loss', 0.11)
#             '''
#             long_update = True
#             if loop_flag == 0:
#                 long_update = False
#             for k, v in d.items():
#                 self.plot(k, v, long_update)

#         def plot_single_win(self, d, win, loop_i=1):
#             """
#             :param d: dict (name, value) i.e. ('loss', 0.11)
#             :param win: only one win
#             :param loop_i: i.e. plot testing loss and label
#             :return:
#             """
#             for k, v in d.items():
#                 x = self.index.get(k, 0)
#                 self.viz.line(Y=np.array([v]), X=np.array([x]),
#                             name=k,
#                             win=win,
#                             opts=dict(title=win, showlegend=True),
#                             update='append' if (x > 0 and loop_i > 0) else None)
#                             # update=None if (x == 0 or loop_i == 0) else 'append')
#                 self.index[k] = x + 1

#         def plot_legend(self, win, name, y, long_update=True, **kwargs):
#             '''
#             plot different line in different time in the same window
#             One mame, one win: only one lie in a win.
#             '''
#             # eg. 
#             # self.vis.plot_legend(win='iou', name='val', y=iou.mean())
#             x = self.index.get(
#                 name, 0)  # dict.get(key, default=None).
#             self.viz.line(Y=np.array([y]), X=np.array([x]),
#                         name=name,
#                         win=win,
#                         opts=dict(title=win, showlegend=True),
#                         update='append' if (x > 0 and long_update) else None,
#                         **kwargs)
#             self.index[name] = x + 1    # Maintain the X

#         def plot(self, name, y, long_update, **kwargs):
#             '''
#             self.plot('loss', 1.00)
#             One mame, one win: only one lie in a win.
#             '''
#             x = self.index.get(
#                 name, 0)  # dict.get(key, default=None).
#             self.viz.line(Y=np.array([y]), X=np.array([x]),
#                         win=name,
#                         opts=dict(title=name),
#                         update='append' if (x > 0 and long_update) else None,
#                         **kwargs)
#             self.index[name] = x + 1    # Maintain the X

#         def draw_roc(self, fpr, tpr):
#             self.viz.line(Y=np.array(tpr), X=np.array(fpr),
#                         name='roc_curve',
#                         win='roc_curve',
#                         opts=dict(title='roc_curve', showlegend=True))
            
#     vis = Visualizer()




from pycamia import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "micomputing",
    author = "Yuncheng Zhou",
    create = "2022-03",
    fileinfo = "File containing U-net and convolutional networks.",
    help = "Use `from micomputing import *`.",
    requires = "batorch"
).check()

__all__ = """
    U_Net
    CNN
    FCN
    RNN
    NeuralODE
    Convolution_Block
    Convolution
    Models
""".split()
    
import math

with __info__:
    import batorch as bt
    from batorch import nn
    from pycamia import touch, avouch, execblock
    from pycamia import Path, ByteSize, SPrint, infinite_itemize

def parse(string):
    if string.count('(') > 1 or string.count(')') > 1: raise TypeError("Invalid to parse: " + string + ". ")
    if string.count('(') == 0 and string.count(')') == 0: string += '()'
    return eval('("' + string.lower().replace('(', '", (').replace(')', ',)').replace('(,)', '()') + ')')

def cat(*tensors): return bt.cat(tensors, 1)
def combine(list_of_items, reduction):
    if len(list_of_items) >= 2:
        z = reduction(list_of_items[0], list_of_items[1])
        for i in range(2, len(list_of_items)):
            z = reduction(z, list_of_items[i])
    else: z = list_of_items[0]
    return z

Convolution = {
    0: lambda ic, oc, *_: nn.Linear(ic, oc), 
    1: nn.Conv1d, 
    2: nn.Conv2d, 
    3: nn.Conv3d
}

MaxPool = {
    0: lambda *_: (lambda x: x), 
    1: nn.MaxPool1d, 
    2: nn.MaxPool2d, 
    3: nn.MaxPool3d
}

ConvTranspose = {
    0: lambda *_: (lambda x: x), 
    1: nn.ConvTranspose1d, 
    2: nn.ConvTranspose2d, 
    3: nn.ConvTranspose3d
}

class BatchNorm(nn.Module):
    def __init__(self, idim, ch):
        super().__init__()
        self.major = nn.BatchNorm1d(ch)
        self.idim = idim
        
    def __class_getitem__(cls, idim):
        if idim in [1, 2, 3]: return getattr(nn, f"BatchNorm{idim}d")
        return lambda ch: cls(idim, ch)
    
    def forward(self, x):
        if self.idim == 0:
            return self.major(x.unsqueeze(-1)).squeeze(-1)
        else: return self.major(x)

class Convolution_Block(nn.Module):
    '''
    Args:
        dimension (int): The dimension of the images. Defaults to 2. 
        in_channels (int): The input channels for the block. 
        out_channels (int): The output channels for the block. 
        mid_channels (int): The middle channels for the block. 
        conv_num (int): The number of convolution layers. Defaults to 1. 
        kernel_size (int): The size of the convolution kernels. Defaults to 3. 
        padding (int): The image padding for the convolutions. Defaults to 1. 
        initializer (str): A string indicating the initialing strategy. Possible values 
            are normal(0, 0.1) or uniform(-0.1, 0.1) or constant(0) (all parameters can be changed)
        linear_layer (bool): Whether the convolutions are used for 1x1 images (equivalent to linear layers). 
        activation_function (class): The activation function. 
        final_activation (class): The activation function after the final convolution. 
        active_args (dict): The arguments for the activation function. 
        conv_block (str): A string with possible values in ('conv', 'dense', 'residual'), 
            indicating which kind of block the U-Net is using: normal convolution layers, 
            DenseBlock or ResidualBlock. Defaults to 'conv'.
        res_type (function): The combining type for the residual connections.
    '''
    
    def __init__(self, in_channels, out_channels, mid_channels=None, **params):
        super().__init__()
        default_values = {'dimension': 2, 'conv_num': 1, 'kernel_size': 3, 'padding': 1, 'linear_layer': False, 'initializer': "normal(0, 0.1)", 'conv_block': 'conv', 'res_type': bt.add, 'activation_function': nn.ReLU, 'final_activation': ..., 'active_args': {}}
        param_values = {}
        param_values.update(default_values)
        param_values.update(params)
        self.__dict__.update(param_values)
        self.in_channels = in_channels
        self.out_channels = out_channels
        if mid_channels is None:
            if self.linear_layer: self.mid_channels = 5 * self.out_channels
            else: self.mid_channels = self.out_channels
        else:
            self.mid_channels = mid_channels
            
        if self.linear_layer: self.kernel_size = 1; self.padding = 0
        
        if isinstance(self.padding, str): self.padding = {'SAME': self.kernel_size // 2, 'ZERO': 0, 'VALID': 0}.get(self.padding.upper(), self.kernel_size // 2)
        if self.activation_function is None: self.activation_function = lambda *a, **k: (lambda x: x)
        if self.final_activation == ...: self.final_activation = self.activation_function
        if self.final_activation is None: self.final_activation = lambda *a, **k: (lambda x: x)
        
        self.layers = nn.ModuleList()
        for i in range(self.conv_num):
            ic = self.in_channels if i == 0 else ((self.mid_channels * i + self.in_channels) if self.conv_block == 'dense' else self.mid_channels)
            oc = self.out_channels if i == self.conv_num - 1 else self.mid_channels
            conv = Convolution[self.dimension](ic, oc, self.kernel_size, 1, self.padding)
            initialize_model, initialize_params = parse(self.initializer)
            eval('nn.init.%s_' % initialize_model)(conv.weight, *initialize_params)
            if self.conv_block != 'dense': self.layers.append(conv)
            oc = (self.mid_channels * i + self.in_channels) if self.conv_block == 'dense' else oc
            if not self.linear_layer: self.layers.append(BatchNorm[self.dimension](oc))
            if i < self.conv_num-1: self.layers.append(self.activation_function(**self.active_args))
            if self.conv_block == 'dense': self.layers.append(conv)

    def forward(self, x):
        self.memory_used = ByteSize(0)
        need_squeeze = False
        original_shape = x.shape
        if self.linear_layer and x.n_space_dim == 0:
            need_squeeze = True
            x = x.view(*x.shape.python_repr, *((1,) * self.dimension))
        if self.conv_block == 'dense':
            conv_results = [x]
            conv_layer = True
            for layer in self.layers:
                try:
                    if conv_layer: x = layer(bt.cat([bt.crop_as(l, conv_results[-1]) for l in conv_results], 1))
                    else: x = layer(x)
                except Exception as e:
                    raise e.__class__(f"In layer {layer}. " + e.__str__())
                self.memory_used += x.byte_size()
                conv_layer = layer.__class__.__name__.startswith('Conv')
                if conv_layer: conv_results.append(x)
            result = self.final_activation(**self.active_args)(x)
        else:
            y = x
            for layer in self.layers:
                try: y = layer(y)
                except Exception as e:
                    raise e.__class__(f"In layer {layer}. " + e.__str__())
                self.memory_used += ByteSize(y.numel() * y.element_size())
            y = y.as_subclass(bt.Tensor).special_from(x)
            if self.conv_block == 'residual': z = self.res_type(bt.crop_as(x, y), y)
            else: z = y
            result = self.final_activation(**self.active_args)(z)
        result = result.as_subclass(bt.Tensor).special_from(x)
        if need_squeeze: result = result.view(original_shape.with_feature(result.feature))
        elif self.padding == self.kernel_size // 2:
            return bt.crop_as(result, x.space)
        return result

class U_Net(nn.Module):
    '''
    Args:
        dimension (int): The dimension of the images. Defaults to 2 (see U-Net). 
        depth (int): The depth of the U-Net. Defaults to 4 indicating 4 pooling layers and 4 up-sampling layers (see U-Net).
        conv_num (int): The number of continuous convolutions in one block. Defaults to 2. 
        padding (int or str): Indicate the type of padding used. Defaults to 'SAME' though it is 0 in conventional U-Net. 
        in_channels (int): The number of channels for the input. Defaults to 1 (see U-Net).
        out_channels (int): The number of channels for the output. Defaults to 2 (see U-Net).
        block_channels (int): The number of channels for the first block if a number is provided. Defaults to 64 (see U-Net). 
            If a list is provided, the length should be the same as the number of blocks plus two (2 * depth + 3). It represents the channels before and after each block (with the output channels included). 
            Or else, a function may be provided to compute the output channels given the block index (-1 ~ 2 * depth + 1) [including input_channels at -1 and output_channels at 2 * depth + 1]. 
        bottleneck_out_channels (int): The number of channels for the bottleneck output. Defaults to 0. 
        kernel_size (int): The size of the convolution kernels. Defaults to 3 (see U-Net). 
        pooling_size (int): The size of the pooling kernels. Defaults to 2 (see U-Net). 
        // keep_prob (float): The keep probability for the dropout layers. 
        conv_block (str): A string with possible values in ('conv', 'dense', 'residual'), indicating which kind of block the U-Net is using: normal convolution layers, DenseBlock or ResidualBlock. 
        multi_arms (str): A string with possible values in ('shared(2)', 'seperate(2)'), indicating which kind of encoder arms are used. 
        multi_arms_combine (function): The combining type for multi-arms. See skip_type for details. 
        skip_type (function): The skip type for the skip connections. Defaults to catenation (cat; see U-Net). Other possible skip types include torch.mul or torch.add. 
        res_type (function): The combining type for the residual connections. It should be torch.add in most occasions. 
        activation_function (class): The activation function used after the convolution layers. Defaults to nn.ReLU. 
        active_args (dict): The arguments for the activation function. Defaults to {}. 
        initializer (str): A string indicating the initialing strategy. Possible values are normal(0, 0.1) or uniform(-0.1, 0.1) or constant(0) (all parameters can be changed)
        with_softmax (bool): Whether a softmax layer is applied at the end of the network. Defaults to True. 
        cum_layers (list): A list consisting two numbers [n, m] indicating that the result would be a summation of the upsamples of the results of the nth to the mth (included) blocks, block_numbers are in range 0 ~ 2 * depth. 
            The negative indices are allowed to indicate the blocks in a inversed order with -1 representing the output for the last block. 
    '''
    
    class Softmax(nn.Module):
        def forward(self, x): return nn.functional.softmax(x, 1)
    
    class Encoder_Block(nn.Module):
        
        def __init__(self, in_channels, out_channels, has_pooling, params):
            super().__init__()
            block_params = params.copy()
            block_params.update({'in_channels': in_channels, 'out_channels': out_channels, 'has_pooling': has_pooling})
            self.__dict__.update(block_params)
            if has_pooling: self.pooling = MaxPool[self.dimension](self.pooling_size, ceil_mode = True)
            self.conv_block = Convolution_Block(**block_params)

        def forward(self, x):
            if self.has_pooling: y = self.pooling(x)
            else: y = x
            return self.conv_block(y)
            
    class Decoder_Block(nn.Module):
        
        def __init__(self, list_of_encoders, in_channels, out_channels, params, copies_of_inputs):
            super().__init__()
            block_params = params.copy()
            block_params.update({'in_channels': in_channels, 'out_channels': out_channels})
            self.__dict__.update(block_params)
            if self.skip_type == cat: to_channels = in_channels - list_of_encoders[0].out_channels
            else: assert all([in_channels == encoder.out_channels for encoder in list_of_encoders]); to_channels = in_channels
            self.upsampling = ConvTranspose[self.dimension](in_channels * copies_of_inputs, to_channels, self.pooling_size, self.pooling_size, 0)
            block_params.update({'in_channels': to_channels + sum([encoder.out_channels for encoder in list_of_encoders]), 'out_channels': out_channels})
            self.conv_block = Convolution_Block(**block_params)

        def forward(self, x, list_of_encoder_results):
            y = self.upsampling(x)
            if self.padding == self.kernel_size // 2:
                to_combine = list_of_encoder_results + [bt.crop_as(y, list_of_encoder_results[0])]
            else: to_combine = [bt.crop_as(encoder_result, y) for encoder_result in list_of_encoder_results] + [y]
            joint = combine(to_combine, self.skip_type)
            return self.conv_block(joint)


    def __init__(self, **params):
        super().__init__()
        default_values = {'dimension': 2, 'depth': 4, 'conv_num': 2, 'padding': 'SAME', 'in_channels': 1, 'out_channels': 2, 'block_channels': 64, 'kernel_size': 3, 'pooling_size': 2, 'keep_prob': 0.5, 'conv_block': 'conv', 'multi_arms': "shared", 'multi_arms_combine': cat, 'skip_type': cat, 'res_type': bt.add, 'activation_function': nn.ReLU, 'active_args': {}, 'initializer': "normal(0, 0.1)", 'with_softmax': True, 'cum_layers': -1, 'bottleneck_out_channels': 0}
        param_values = {}
        param_values.update(default_values)
        param_values.update(params)
        self.__dict__.update(param_values)
        
        if isinstance(self.block_channels, int):
            self.block_channels = [self.in_channels] + [self.block_channels << min(i, 2 * self.depth - i) for i in range(2 * self.depth + 1)] + [self.out_channels]
        bchannels = self.block_channels
        if not callable(self.block_channels): self.block_channels = lambda i: bchannels[i + 1]
        
        if isinstance(self.padding, str): self.padding = {'SAME': self.kernel_size // 2, 'ZERO': 0, 'VALID': 0}.get(self.padding.upper(), self.kernel_size // 2)
        
        if isinstance(self.cum_layers, int): self.cum_layers = [self.cum_layers, self.cum_layers]
        l, u = self.cum_layers
        l = (l + 2 * self.depth + 1) % (2 * self.depth + 1)
        u = (u + 2 * self.depth + 1) % (2 * self.depth + 1)
        if l > u: l, u = u, l
        self.cum_layers = [max(l, self.depth), min(u, 2 * self.depth)]
        
        param_values = {k: self.__dict__[k] for k in param_values}
        
        self.arm_type, self.arm_num = parse(self.multi_arms)
        self.arm_num = 1 if len(self.arm_num) == 0 else self.arm_num[0]
        if self.arm_type == 'shared': self.dif_arm_num = 1
        else: self.dif_arm_num = self.arm_num
        
        for iarm in range(self.dif_arm_num):
            for k in range(self.depth + 1):
                setattr(self, 'block%d_%d' % (k, iarm), self.Encoder_Block(self.block_channels(k - 1), self.block_channels(k), k != 0, param_values))
        
        if self.bottleneck_out_channels > 0:
            param_values = {k: v for k, v in param_values.items() if k not in ('in_channels', 'out_channels')}
            setattr(self, 'bottleneck_out', Convolution_Block(self.block_channels(self.depth) * self.arm_num, self.bottleneck_out_channels, **param_values))

        for k in range(self.cum_layers[0], self.depth + 1):
            conv = Convolution[self.dimension](self.block_channels(k), self.block_channels(2 * self.depth + 1), 1, 1, 0)
            initialize_model, initialize_params = parse(self.initializer)
            eval('nn.init.%s_' % initialize_model)(conv.weight, *initialize_params)
            if k < self.cum_layers[1]:
                setattr(self, 'block%dout' % k, nn.Sequential(conv, self.activation_function(**self.active_args)))
                setattr(self, 'out%dupsample' % k, ConvTranspose[self.dimension](
                    self.block_channels(2 * self.depth + 1), self.block_channels(2 * self.depth + 1), self.pooling_size, self.pooling_size, 0
                ))
            else: setattr(self, 'block%dout' % k, conv)

        for k in range(self.depth + 1, self.cum_layers[1] + 1):
            setattr(self, 'block%d' % k, self.Decoder_Block(
                [getattr(self, 'block%d_%d' % (2 * self.depth - k, iarm)) for iarm in range(self.dif_arm_num)] * (self.arm_num // self.dif_arm_num), 
                self.block_channels(k - 1), self.block_channels(k), param_values, 
                self.arm_num if k == self.depth + 1 and self.multi_arms_combine == cat else 1
            ))
            conv = Convolution[self.dimension](self.block_channels(k), self.block_channels(2 * self.depth + 1), 1, 1, 0)
            initialize_model, initialize_params = parse(self.initializer)
            eval('nn.init.%s_' % initialize_model)(conv.weight, *initialize_params)
            if k < self.cum_layers[1]:
                setattr(self, 'block%dout' % k, nn.Sequential(conv, self.activation_function(**self.active_args)))
                setattr(self, 'out%dupsample' % k, ConvTranspose[self.dimension](
                    self.block_channels(2 * self.depth + 1), self.block_channels(2 * self.depth + 1), self.pooling_size, self.pooling_size, 0
                ))
            else: setattr(self, 'block%dout' % k, conv)

        if self.with_softmax: self.softmax = self.Softmax()
        self.to(bt.get_device().main_device)
        
    @property
    def bottleneck(self):
        if self.bottleneck_out_channels > 0:
            result = getattr(self, f'block{self.depth}result', None)
            if result is None: return
            return self.bottleneck_out(result).mean(...)
        else: return

    def forward(self, x):
        size = x.size()[1:]
        if len(size) == self.dimension and self.in_channels == 1: x = x.unsqueeze(1)
        elif len(size) == self.dimension + 1 and self.in_channels * self.arm_num == size[0]: pass
        else: raise ValueError(f"The input tensor does not correspond to the U-Net structure: got {size}, but requires ([{self.in_channels * self.arm_num}], n_1, ⋯, n_{self.dimension}). ")
        
        assert size[0] % self.arm_num == 0
        inputs = x.split(size[0] // self.arm_num, 1)
        assert len(inputs) == self.arm_num
        
        for i, y in enumerate(inputs):
            for k in range(self.depth + 1):
                y = getattr(self, 'block%d_%d' % (k, 0 if self.arm_type == 'shared' else i))(y)
                setattr(self, 'block%d_%dresult' % (k, i), y)
        
        to_combine = [getattr(self, 'block%d_%dresult' % (self.depth, i)) for i in range(self.arm_num)]
        z = combine(to_combine, self.multi_arms_combine)
        setattr(self, 'block%dresult' % self.depth, z)
        
        for k in range(self.depth + 1, self.cum_layers[1] + 1):
            z = getattr(self, 'block%d' % k)(z, [getattr(self, 'block%d_%dresult' % (2 * self.depth - k, iarm)) for iarm in range(self.arm_num)])
            setattr(self, 'block%dresult' % k, z)

        t = 0
        for k in range(self.cum_layers[0], self.cum_layers[1] + 1):
            setattr(self, 'block_out%dresult' % k, getattr(self, 'block%dout' % k)(getattr(self, 'block%dresult' % k)) + t)
            if k < self.cum_layers[1]: t = getattr(self, 'out%dupsample' % k)(getattr(self, 'block_out%dresult' % k))
        
        if self.with_softmax: return self.softmax(getattr(self, 'block_out%dresult' % k))
        else: return getattr(self, 'block_out%dresult' % k)
        
    def optimizer(self, lr=0.001): return bt.Optimization(bt.optim.Adam, self.parameters(), lr)

    def loss(self, x, y):
        y_hat = self(x)
        clamped = y_hat.clamp(1e-10, 1.0)
        self.y_hat = y_hat
        return - bt.sum(y * bt.log(clamped), 1).mean().mean()
        
    def __getitem__(self, i):
        if self.arm_num == 1 and i <= self.depth: i = (i, 0)
        return getattr(self, 'block%dresult' % i if isinstance(i, int) else 'block%d_%dresult' % i)
        
    def __iter__(self):
        for i in range(2 * self.depth + 1):
            if i <= self.depth:
                for iarm in range(self.arm_num):
                    yield 'block%d_%dresult' % (i, iarm), (i, iarm)
            else: yield 'block%dresult' % i, i

class CNN(U_Net):
    '''
    Args:
        dimension (int): The dimension of the images. Defaults to 2 (see VGG). 
        blocks (int): The number of the downsampling blocks. Defaults to 5 blocks (see VGG).
        conv_num (int or list of length 'blocks'): The number of continuous convolutions in one block. Defaults to [2, 2, 3, 3, 3] (see VGG).
            If the numbers for all blocks are the same, one can use one integer.
        padding (int or str): Indicate the type of padding used. Defaults to 'SAME' indicating a same output shape as the input. 
        in_channels (int): The number of channels for the input. Defaults to 1 (see VGG).
        out_elements (int): The number of channels for the output, as the number of classification. Defaults to 1000 for 1000 classes.
        layer_channels (int or list of length 'blocks'): The number of channels for each block. Defaults to [64, 128, 256, 512, 512] (see VGG). 
            Or else, a function may be provided to compute the output channels given the block index (-1 ~ 2 * depth + 1). 
        kernel_size (int): The size of the convolution kernels. Defaults to 3 (see VGG). 
        pooling_size (int): The size of the pooling kernels. Defaults to 2 (see VGG). 
        // keep_prob (float): The keep probability for the dropout layers. 
        conv_block (str): A string with possible values in ('conv', 'dense', 'residual'), indicating which kind of block the U-Net is using: normal convolution layers, DenseBlock or ResidualBlock. 
        multi_arms (str): A string with possible values in ('shared(2)', 'seperate(2)'), indicating which kind of encoder arms are used. 
        multi_arms_combine (function): The combining type for multi-arms. Defaults to catenation (cat). Other possible skip types include torch.mul or torch.add. 
        res_type (function): The combining type for the residual connections. It should be torch.add in most occasions. 
        activation_function (class): The activation function used after the convolution layers. Defaults to nn.ReLU. 
        active_args (dict): The arguments for the activation function. Defaults to {}. 
        initializer (str): A string indicating the initialing strategy. Possible values are normal(0, 0.1) or uniform(-0.1, 0.1) or constant(0) (all parameters can be changed)
        with_softmax (bool): Whether a softmax layer is applied at the end of the network. Defaults to True. 
    '''
    
    def __init__(self, dimension = 2, blocks = 5, conv_num = 2, padding = 'SAME', 
        in_channels = 1, out_elements = 2, layer_channels = 64, kernel_size = 3, 
        pooling_size = 2, keep_prob = 0.5, conv_block = 'conv', multi_arms = "shared", 
        multi_arms_combine = cat, res_type = bt.add, activation_function = nn.ReLU,
        active_args = {}, initializer = "normal(0, 0.1)", with_softmax = True):
        depth = blocks - 1
        if isinstance(layer_channels, int):
            maxlc = layer_channels
            layer_channels = [in_channels]
            multiplier = math.pow(maxlc / in_channels, 1 / (depth + 1))
            for i in range(depth):
                layer_channels.append(int(layer_channels[-1] * multiplier))
            layer_channels.append(maxlc)
            layer_channels.extend([0] * depth)
            layer_channels.append(out_elements)
        super().__init__(dimension = dimension, depth = depth, conv_num = conv_num, 
            padding = padding, in_channels = in_channels, out_channels = out_elements, 
            block_channels = layer_channels, kernel_size = kernel_size, 
            pooling_size = pooling_size, keep_prob = keep_prob, conv_block = conv_block,
            multi_arms = multi_arms, multi_arms_combine = multi_arms_combine, skip_type = None,
            res_type = res_type, activation_function = activation_function, active_args = active_args,
            initializer = initializer, with_softmax = with_softmax, cum_layers = depth)

    def forward(self, x):
        wsm = self.with_softmax
        self.with_softmax = False
        if wsm: r = self.softmax(super().forward(x).flatten(2).mean(-1))
        else: r = super().forward(x).flatten(2).mean(-1)
        self.with_softmax = wsm
        return r
        
class FCN(nn.Module):
    '''
    Fully connected network, with hidden layers of increased and then decreased sizes. 
        For layer_elements = 64 and layers = 8 and in_elements = out_elements = 8, 
        the layer sizes are [8, 16, 32, 64, 64, 32, 16, 8]. 
    
    Args:
        layers (int): Indicate the number of fully connected layers. 
        in_elements (int): The number of elements for the input. Defaults to 1.
        out_elements (int): The number of elements for the output, as the number of classification. Defaults to 1000 for 1000 classes.
        layer_elements (int or list of length 'layers'): The number of channels for each block. In a VGG, it should be [64, 128, 256, 512, 512]. 
            Or else, a function may be provided to compute the output channels given the block index (-1 ~ 2 * depth + 1). 
        kernel_size (int): The size of the convolution kernels. Defaults to 3. 
        keep_prob (float): The keep probability for the dropout layers. 
        activation_function (class): The activation function used after the convolution layers. Defaults to nn.ReLU. 
        active_args (dict): The arguments for the activation function. Defaults to {}. 
        initializer (str): A string indicating the initialing strategy. Possible values are normal(0, 0.1) or uniform(-0.1, 0.1) or constant(0) (all parameters can be changed)
        with_softmax (bool): Whether a softmax layer is applied at the end of the network. Defaults to True. 
    '''

    class Softmax(nn.Module):
        def forward(self, x): return nn.functional.softmax(x, 1)
    def __init__(self, layers = 4, in_elements = 1, out_elements = 2, layer_elements = 64, 
        keep_prob = 0.5, activation_function = nn.ReLU, active_args = {}, 
        initializer = "normal(0, 0.1)", with_softmax = True):
        if isinstance(layer_elements, int):
            maxlc = layer_elements
            layer_elements = [in_elements]
            multiplier = bt.pow(maxlc / in_elements, 1 / (layers // 2 - 1))
            for i in range(layers // 2 - 1):
                layer_elements.append(int(layer_elements[-1] * multiplier))
            layer_elements.append(maxlc)
            if layers % 2 == 0: layer_elements.extend(layer_elements[-2::-1])
            else: layer_elements.extend(layer_elements[::-1])
            layer_elements[-1] = out_elements
        if isinstance(layer_elements, list):
            lc = layer_elements.copy()
            layer_elements = lambda i: lc[i]
        self.layers = []
        for l in range(layers):
            fcl = nn.Linear(layer_elements(l), layer_elements(l+1))
            initialize_model, initialize_params = parse(initializer)
            eval('nn.init.%s_' % initialize_model)(fcl.weight, *initialize_params)
            self.layers.append(fcl)
            if l < layers - 1:
                self.layers.append(activation_function(**active_args))
                self.layers.append(nn.Dropout(keep_prob))
            elif with_softmax:
                self.layers.append(self.Softmax())
        self.struct = nn.Sequential(*self.layers)
        self.to(bt.get_device().main_device)

    def forward(self, x):
        return self.struct(x)


class MultiVariate(list):
    def __init__(self, *args, main_index=0, main_grad_index=1):
        if len(args) > 1: args = (list(args),)
        super().__init__(*args)
        self.main_index = main_index
        self.main_grad_index = main_grad_index
        
    for op in """
        __add__ __iadd__ __radd__
        __sub__ __isub__ __rsub__
        __mul__ __imul__ __rmul__
        __div__ __idiv__ __rdiv__
        __truediv__ __itruediv__ __rtruediv__
        __floordiv__ __ifloordiv__ __rfloordiv__""".split():
        execblock(f"""
        def {op}(self, other):
            if isinstance(other, MultiVariate):
                return MultiVariate((getattr(a, '{op}')(b) for a, b in zip(self, other)), main_index=self.main_index, main_grad_index=self.main_grad_index)
            return MultiVariate((getattr(a, '{op}')(other) for a in self), main_index=self.main_index, main_grad_index=self.main_grad_index)
        """)
            
    for op in """
        abs norm max min item detach""".split():
        execblock(f"""
        def {op}(self): return MultiVariate((a.{op}() for a in self), main_index=self.main_index, main_grad_index=self.main_grad_index)
        """)
            
    for prop in """
        shape""".split():
        execblock(f"""
        @property
        def {prop}(self): return [a.{prop} for a in self]
        """)
        
    def __str__(self): return ' : '.join([str(a.round(decimals=4).tolist()) for a in self])
            
    def cat(self): return bt.cat([a.flatten() for a in self])
    
    def max_item(self): return bt.tensor(self.max().item()).max()
        
    @property
    def main(self): return self[self.main_index]
        
    @main.setter
    def main(self, value): self[self.main_index] = value
        
    @property
    def main_grad(self): return self[self.main_grad_index]
        
    @main_grad.setter
    def main_grad(self, value): self[self.main_grad_index] = value

def multi_variate_max(mv_x, mv_y):
    return MultiVariate((bt.max(x, y) for x, y in zip(mv_x, mv_y)), main_index=mv_x.main_index, main_grad_index=mv_x.main_grad_index)


class NeuralODE(nn.Module):
    '''
    Neural ODE structue. 
    => [main stream: inputs] main_channels
    => [coupled stream: reference maps] coupled_channels
    
    UNODE: unsupervised ODE network, the standard ODE framework.                coupled_channels = 0
    SNODE: supervised ODE, by inserting the features to the ODE framework.      coupled_channels > 0; coupled_ODE = False
    CNODE: coupled ODE, supervise the ODE framework by a coupled ODE sequence.  coupled_channels > 0; coupled_ODE = True
    R-NODE: NODE versions with regularization.                                  regularization != None
    
    Args:
        dimension (int): The dimension of the images. 
        layers (int): The number of accumulated layers, -1 means automatic number of layers. 
            When layers = -1, we have additional arguments:
                time_start (float): The starting time. Defaults to 0. 
                frequency (int): The time frequency for marching, that is the time spot for the i-th iteration when frequency=50Hz is i/(frequency=50).
                max_layers (int): The maximal number of marches. 
                rtol (float): The relative tolerance for auto layer and stopping criteria. 
                atol (float): The absolute tolerance for auto layer and stopping criteria. Together with `rtol`, they indicates a tolerance of `rtol x magnitute + atol` error for the gradient estimation before ending the algorithm. 
        method (str): A string with possible values in ('Euler', 'RK2', 'RK4'), indicating the gradient estimation method. Defaults to 'Euler'. 
        conv_num (int): The number of continuous convolutions in each layer. Defaults to 2. 
        main_channels (int): The number of channels for the main stream ODE. Defaults to 2.
        coupled_channels (int): The number of channels for the coupled stream ODE. Defaults to 0, indicating no coupled network. 
        mid_channels (int): The number of channels for the hidden layers in main stream. Defaults to 10.
        main_march_func (func): The gradient estimator of the main stream ODE. Defaults to a simple stack of `conv_num` convolution layers. 
        coupled_march_func (func): The gradient estimator of the coupled stream ODE. Defaults also to the simple stack of `conv_num` convolution layers. 
        time_points (func): The function that returns the time point at iteration i. Defaults to the linear function (with a time frequency of `frequency`). 
        auto_time_step (bool): Whether to auto decide the time step. `time_points` is disabled when it is True. Defaults to True. 
        min_step (float): The minimal step size in auto adaptive steps. 
        kernel_size (int): The size of the convolution kernels. Defaults to 3. 
        padding (int or str): Indicate the type of padding used. Defaults to 'SAME' indicating a same output shape as the input. 
        conv_block (str): A string with possible values in ('conv', 'dense', 'residual'), indicating which kind of block for the layers: normal convolution layers, DenseBlock or ResidualBlock. 
        res_type (function): The combining type for the residual connections. It should be torch.add in most occasions. 
        activation_function (class): The activation function used after the convolution layers. Defaults to nn.ReLU. 
        final_activation (class): The activation function after the final convolution. Defaults to self.activation_function. 
        active_args (dict): The arguments for the activation function. Defaults to {}. 
        initializer (str): A string indicating the initialing strategy. Possible values are normal(0, 0.1) or uniform(-0.1, 0.1) or constant(0) (all parameters can be changed)
        regularization (func): The regularization term for the delta terms. 
            The regularization function which has the arguments (i, t, dt, *[main_x, coupled_x (if used)], delta_main_x)
    '''
    
    class Softmax(nn.Module):
        def forward(self, x): return nn.functional.softmax(x, 1)
    
    def __init__(self, dimension = 2, layers = -1, method = 'Euler', conv_num = 2, kernel_size = 3, padding = 1, 
        time_start = 0., time_end = 1., frequency = 50, rtol = 1e-3, atol = 1e-3, max_layers = 1000, linear_layer = False, min_step = 0, 
        coupled_channels = 0, main_channels = 2, mid_channels = None, conv_block = 'conv', res_type = bt.add, 
        main_march_func = None, coupled_march_func = None, time_points = None, auto_time_step = True,
        activation_function = nn.ReLU, final_activation = ..., active_args = {}, initializer = "normal(0, 0.1)", 
        shared = True, coupled_ODE = True, time_channels = 1, auto_grad = False, regularization = None):
        super().__init__()
        
        if coupled_channels == True: coupled_channels = 64
        if mid_channels is None: mid_channels = main_channels
        
        params = dict(dimension=dimension, mid_channels=mid_channels, linear_layer=linear_layer, conv_num=conv_num, kernel_size=kernel_size, padding=padding, 
            activation_function=activation_function, final_activation=final_activation, active_args=active_args, conv_block=conv_block, res_type=res_type, initializer=initializer)
        avouch(layers >= 0 or shared, "Unshared Neural ODE network should have fixed number of layers. ")
        if coupled_channels > 0:
            if coupled_ODE:
                if shared:
                    if coupled_march_func is not None: self.coupled_march_func = coupled_march_func
                    else: self.coupled_march_func = Convolution_Block(in_channels=coupled_channels + time_channels, out_channels=coupled_channels, **params)
                    self.coupled_march = infinite_itemize(self.coupled_march_func, n=layers)
                else:
                    self.coupled_march = []
                    for i in range(layers):
                        if coupled_march_func is not None: setattr(self, f'coupled_march_func_{i}', coupled_march_func(i))
                        else: setattr(self, f'coupled_march_func_{i}', Convolution_Block(in_channels=coupled_channels + time_channels, out_channels=coupled_channels, **params))
                        self.coupled_march.append(getattr(self, f'coupled_march_func_{i}'))
        if shared:
            if main_march_func is not None: self.main_march_func = main_march_func
            else: self.main_march_func = Convolution_Block(in_channels=main_channels + coupled_channels + time_channels, out_channels=main_channels, **params)
            self.main_march = infinite_itemize(self.main_march_func, n=layers)
        else:
            self.main_march = []
            for i in range(layers):
                if main_march_func is not None: setattr(self, f'main_march_func_{i}', main_march_func(i))
                else: setattr(self, f'main_march_func_{i}', Convolution_Block(in_channels=main_channels + coupled_channels + time_channels, out_channels=main_channels, **params))
                self.main_march.append(getattr(self, f'main_march_func_{i}'))
        self.method = method
        self.layers = layers
        self.time_start = time_start
        self.time_end = time_end
        self.frequency = layers if layers >= 0 else frequency
        self.rtol = rtol
        self.atol = atol
        self.max_layers = max_layers
        self.shared = shared
        self.coupled_ODE = coupled_ODE
        self.auto_grad = auto_grad
        self.coupled_channels = coupled_channels
        self.time_channels = time_channels
        self.regularization = regularization
        
        if time_points is None:
            if layers >= 0: self.time_end = self.time_start + layers / self.frequency
            time_points = bt.arange(int((self.time_end - self.time_start) * self.frequency + 1)).float() / self.frequency + self.time_start
        self.time_points = time_points
        self.auto_time_step = auto_time_step
        self.min_step = min_step
    
    def time(self, t, n_channel):
        return t * bt.ones(self.reference.shape.with_feature((n_channel,)), dtype=self.reference.dtype, device=self.reference.device)
    
    def error_ratio(self, x0, x1=None, dx=None):
        if x1 is not None and dx is not None or x1 is None and dx is None:
            raise TypeError("Function tolerant takes arguments (x0, x1) or (x0, dx). ")
        elif x1 is None: x1 = x0 + dx
        else: dx = x1 - x0
        error = dx.abs().max()
        tolerance = self.rtol * multi_variate_max(x0.abs().max(), x1.abs().max()).item() + self.atol
        return bt.tensor_to((error / tolerance).item(), x0[0]).max()
    
    def integrate(self, init_x: MultiVariate, diff_func, ts, t0 = None, time_inv = False, steps = None, output_sequence=False, 
                  order=0.9, min_factor=0.2, max_factor=10.): # auto step size parameters
        self.memory_used = ByteSize(0)
        x = init_x
        if steps is None and self.layers >= 0: steps = self.layers
        ts = bt.tensor_to(ts, x[0])
        if t0 is None: t0 = ts[0]
        if not isinstance(t0, bt.Tensor): t0 = bt.tensor(t0)
        
        i_iter = 0
        cur_time_point = t0
        cur_time_step = bt.tensor_to(1. / self.frequency, x[0])
        i_coming_t1 = 0
        coming_t1 = ts[0]
        
        self.log = SPrint()
        if time_inv: self.log(f": Integrating an inversed process...")
        else:
            self.log(f": Integrating a forward process...")
            self.output_sequence = []
        
        while True:
            if not self.auto_grad: self.memory_used = ByteSize(0)
            if time_inv:
                if i_iter >= self.n_iter:
                    self.log("# Stopping the iterations: scheduled steps reached. ")
                    break
                i = int(self.n_iter - i_iter - 1)
            else:
                if steps is not None and i_iter >= steps:
                    self.log("# Stopping the iterations: scheduled steps reached. ")
                    break
                i = i_iter

            t = cur_time_point
            dt = cur_time_step
            
            if bt.equals(coming_t1, t):
                i_coming_t1 += 1
                complete = i_coming_t1 >= len(ts)
                if output_sequence or complete:
                    self.log(f"* Recording sequence value (fetching sequential grad) at time {t}. ")
                    if not time_inv: # In forward process. 
                        self.output_sequence.append(x.main)
                    else: # In backward process: only when self.auto_grad = False. 
                        cur_grad = None if self.output_sequence_grad is None else (self.ouput_sequence_grad[i] if i < len(self.output_sequence_grad) else None)
                        # if cur_grad is not None: x.main_grad = x.main_grad + cur_grad
                if complete: break
                coming_t1 = ts[i_coming_t1]
            if coming_t1 < t + dt:
                dt = coming_t1 - t
            
            self.log(f"> Iteration {i} at time {t}(+{dt}), targeting {coming_t1}. ")
            if self.auto_time_step:
                for _ in range(3): # The maximal time-out number of iteration. 
                    # Adjust the step size using Runge-Kutta Adaptive Step-Size
                    method_ref = "RK2" if self.method != "RK2" else "RK4"
                    grad = diff_func(i, t, dt, x)
                    grad_ref = diff_func(i, t, dt, x, method = method_ref)
                    error_ratio = self.error_ratio(grad_ref, grad) # error / tolerance
                    if error_ratio > 1: # reduce dt when out of tolerance
                        self.log(f"\t| Reducing dt when error rate {error_ratio} > 1 for grad({self.method})={grad} and reference({method_ref})={grad_ref}")
                        factor = 0.9 * error_ratio ** (-order) # 0.9 is a safety setting to avoid local minimum and underflow. 
                        factor = min(max(factor, min_factor), max_factor) # clamp the factor, within [0.2, 10] by default. 
                        cur_time_step = (cur_time_step * factor).clamp(self.min_step)
                        dt = cur_time_step
                    else: break
                else: self.log(f"! Warning: auto time step time out at {t}(+{dt}).")
                cur_time_point = t + dt
            else:
                grad = diff_func(i, t, dt, x)
            
            dx = dt * grad
            error_ratio = self.error_ratio(x, dx=dx)
            if steps is None and (error_ratio < 1 or i_iter >= self.max_layers):
                if i_iter >= self.max_layers: self.log(f"# Stopping the iterations: maximal steps ({self.max_layers}) reached. ")
                else: self.log(f"# Stopping the iterations: converged (with an error ratio of {error_ratio} for x_t={x} and grad={grad}). ")
                if i_iter < 5: print(f"Warning: only {i_iter} NFEs for the ODE. ")
                if i_iter >= self.max_layers: print(f"Warning: reached {self.max_layers} NFE in ODE. ")
                break
            self.log(f". Accept step with an error ratio of {error_ratio} for x_t={x} and grad={grad}). ")
            x = (x - dx) if time_inv else (x + dx)
            
            i_iter += 1
        self.n_iter = i_iter
        return x
    
    # The major marching function
    def march(self, i, t, dt, x, march_funcs, method=None):
        avouch(dt > 0, f'underflow in dt {dt}')
        avouch(bt.isfinite(x).all(), f'infinite x: {x}')
        
        # Select the marching function and the time point for iteration i. 
        if isinstance(march_funcs, infinite_itemize) and march_funcs.is_infinite:
            march_func = march_funcs.obj
        elif i < len(march_funcs): march_func = march_funcs[int(i)]
        else: march_func = march_funcs.obj
        
        # Add the time tensor if necessary. 
        if self.time_channels > 0:
            step = lambda t, x: march_func(bt.cat(x, self.time(t, n_channel=self.time_channels), []))
        else: step = lambda t, x: march_func(x)

        # Solve the ODE using numerical methods. 
        if method is None: method = self.method
        if method == 'Euler':
            grad = step(t, x)
            self.memory_used += getattr(march_func, 'memory_used', 0)
        elif method.startswith('RK'):
            rank = int(method[2:])
            butcher_table = [
                # Order 1: Euler
                [[0     , 0     ], 
                 [0     , 1     ]], 
                # Order 2: Heun Method
                [[0     , 0     , 0     ], 
                 [1     , 1     , 0     ], 
                 [0     , 1 / 2 , 1 / 2 ]],
                # Order 3: 
                [[0     , 0     , 0     , 0     ], 
                 [1     , 1     , 0     , 0     ], 
                 [1 / 2 , 1 / 4 , 1 / 4 , 0     ], 
                 [0     , 1 / 6 , 1 / 6 , 2 / 3 ]], 
                # Order 4
                [[0     , 0     , 0     , 0     , 0     ], 
                 [1 / 2 , 1 / 2 , 0     , 0     , 0     ], 
                 [1 / 2 , 0     , 1 / 2 , 0     , 0     ], 
                 [1     , 0     , 0     , 1     , 0     ], 
                 [0     , 1 / 6 , 1 / 3 , 1 / 3 , 1 / 6 ]], 
                # Order 5
                [[0     , 0         , 0         , 0         , 0             , 0         , 0     ], 
                 [1 / 4 , 1 / 4     , 0         , 0         , 0             , 0         , 0     ], 
                 [3 / 8 , 3 / 32    , 9 / 32    , 0         , 0             , 0         , 0     ], 
                 [12/13 , 1932/2197 , -7200/2197, 7296/2197 , 0             , 0         , 0     ], 
                 [1     , 439 / 216 , -8        , 3680/513  , -845 / 4104   , 0         , 0     ], 
                 [1 / 2 , -8 / 27   , 2         , -3544/2565, 1859 / 4104   , -11 / 40  , 0     ], 
                 [0     , 16 / 135  , 0         , 6656/12825, 28561 / 56430 , -9 / 50   , 2 / 55]]
            ][rank]
            num_eq = len(butcher_table) - 1
            Ks = []
            for i in range(num_eq + 1):
                ci, *ai_ = butcher_table[i]
                if len(Ks) == 0: grad = bt.zeros_like(x[0])
                else: grad = bt.tensor_to(bt.stack([a * k for a, k in zip(ai_, Ks)], bt.func_dim).sum(bt.func_dim), x[0])
                if i == num_eq: break
                Ks.append(step(t + ci * dt, x + dt * grad))
            self.memory_used += num_eq * getattr(march_func, 'memory_used', 0)
        
        return grad
    
    def forward_diff_func(self, i, t, dt, x, method=None):
        if self.coupled_channels > 0:
            main_x, coupled_x = x
            grad_main_x = self.march(i, t, dt, bt.cat(main_x, coupled_x, []), self.main_march, method=method)
            if self.coupled_ODE: # CNODE
                grad_coupled_x = self.march(i, t, dt, coupled_x, self.coupled_march, method=method)
            else: grad_coupled_x = bt.zeros_like(coupled_x)
            return MultiVariate([grad_main_x, grad_coupled_x]) # SNODE / CNODE
        return MultiVariate([self.march(i, t, dt, x[0], self.main_march, method=method)]) # UNODE
    
    def forward_func(self, init_x, coupled_init_x=None, steps=None, output_sequence=False):
        self.reference = init_x
        self.output_sequence = []
        if self.coupled_channels > 0:
            self.output, self.coupled = self.integrate(MultiVariate([init_x, coupled_init_x]), self.forward_diff_func, self.time_points, steps=steps, output_sequence=output_sequence)
        else: self.output, = self.integrate(MultiVariate([init_x]), self.forward_diff_func, self.time_points, steps=steps, output_sequence=output_sequence)
        if output_sequence:
            return self.output, self.output_sequence
        else: return self.output
    
    def UNODE_backward_diff_func(self, i, t, dt, x, method=None):
        main_x, d_main = x
        with bt.enable_grad():
            main_x = main_x.detach().requires_grad_(True)
            delta_main_x = self.march(i, t, dt, main_x, self.main_march, method=method)
            reg_grad_delta = 0
            if self.regularization is not None:
                reg_delta_main_x = delta_main_x.detach().requires_grad_(True)
                reg = self.regularization(i, t, dt, main_x, reg_delta_main_x)
                if reg is not None and reg.requires_grad:
                    reg.backward()
                    reg_grad_main = reg_delta_main_x.grad
                    if reg_grad_main is not None: reg_grad_delta = reg_grad_main
            delta_main_x.backward(d_main.detach() + reg_grad_delta)
        return MultiVariate([delta_main_x.detach(), -main_x.grad])
    
    def SNODE_backward_diff_func(self, i, t, dt, x):
        main_x, d_main, d_coupled = x
        with bt.enable_grad():
            main_x = main_x.detach().requires_grad_(True)
            coupled_x = self.coupled.detach().requires_grad_(True)
            delta_main_x = self.march(i, t, dt, bt.cat(main_x, coupled_x, []), self.main_march, method=method)
            reg_grad_delta = 0
            if self.regularization is not None:
                reg_delta_main_x = delta_main_x.detach().requires_grad_(True)
                reg = self.regularization(i, t, dt, main_x, coupled_x, reg_delta_main_x)
                if reg is not None and reg.requires_grad:
                    reg.backward()
                    reg_grad_main = reg_delta_main_x.grad
                    if reg_grad_main is not None: reg_grad_delta = reg_grad_main
            delta_main_x.backward(d_main.detach() + reg_grad_delta)
        return MultiVariate([delta_main_x.detach(), -main_x.grad, -coupled_x.grad])
    
    def CNODE_backward_diff_func(self, i, t, dt, x):
        main_x, coupled_x, d_main, d_coupled = x
        with bt.enable_grad():
            main_x = main_x.detach().requires_grad_(True)
            coupled_x = coupled_x.detach().requires_grad_(True)
            delta_main_x = self.march(i, t, dt, bt.cat(main_x, coupled_x, []), self.main_march, method=method)
            delta_coupled_x = self.march(i, t, dt, coupled_x, self.coupled_march, method=method)
            reg_grad_delta = 0
            if self.regularization is not None:
                reg_delta_main_x = delta_main_x.detach().requires_grad_(True)
                reg = self.regularization(i, t, dt, main_x, coupled_x, reg_delta_main_x)
                if reg is not None and reg.requires_grad:
                    reg.backward()
                    reg_grad_main = reg_delta_main_x.grad
                    if reg_grad_main is not None: reg_grad_delta = reg_grad_main
            delta_coupled_x.backward(d_coupled.detach())
            delta_main_x.backward(d_main.detach() + reg_grad_delta)
        return MultiVariate([delta_main_x.detach(), delta_coupled_x.detach(), -main_x.grad, -coupled_x.grad], main_grad_index=2)
    
    def backward(self, output_grad):
        output_grad = output_grad.as_subclass(bt.Tensor).view(self.output.shape)
        inits = []
        if self.coupled_channels > 0:
            if self.coupled_ODE:
                inits = MultiVariate([self.output, self.coupled, output_grad, bt.zeros_like(self.coupled)], main_grad_index=2)
                ret = self.integrate(inits, self.CNODE_backward_diff_func, self.time_points[::-1], output_sequence=self.use_output_sequence)[2:]
            else:
                inits = MultiVariate([self.output, output_grad, bt.zeros_like(self.coupled)])
                ret = self.integrate(inits, self.SNODE_backward_diff_func, self.time_points[::-1], output_sequence=self.use_output_sequence)[1:]
        else:
            inits = MultiVariate([self.output, output_grad])
            ret = self.integrate(inits, self.UNODE_backward_diff_func, self.time_points[::-1], output_sequence=self.use_output_sequence)[1:]
        # for mlayer in ([self.main_march[0]] if self.shared else self.main_march):
        #     for p in mlayer.parameters(): p.grad = -p.grad
        # if self.coupled_channels > 0:
        #     for mlayer in ([self.coupled_march[0]] if self.shared else self.coupled_march):
        #         for p in mlayer.parameters(): p.grad = -p.grad
        return ret
    
    class Autograd_Func(bt.autograd.Function):
        @staticmethod
        def forward(ctx, self, init_x, coupled_init_x=None):
            ctx.self = self
            with bt.no_grad():
                return self.forward_func(init_x, coupled_init_x=coupled_init_x, steps=self.n_steps, output_sequence=self.use_output_sequence)
        
        @staticmethod
        def backward(ctx, output_grad, output_sequence_grad = None):
            ctx.self.output_sequence_grad = output_sequence_grad
            return None, *ctx.self.backward(output_grad), None
    
    def forward(self, init_x, coupled_init_x=None, steps=None, output_sequence=False):
        if self.auto_grad: return self.forward_func(init_x, coupled_init_x=coupled_init_x, steps=steps, output_sequence=output_sequence)
        x_has_grad = init_x.requires_grad
        if not init_x.requires_grad: init_x.requires_grad = True
        self.n_steps = steps
        self.use_output_sequence = output_sequence
        if self.coupled_channels == 0: ret = self.Autograd_Func.apply(self, init_x)
        else: ret = self.Autograd_Func.apply(self, init_x, coupled_init_x)
        if not x_has_grad: init_x.grad = None; init_x.requires_grad = False
        def convert(r):
            if isinstance(r, (tuple, list)):
                return r.__class__(convert(s) for s in r)
            r = r.as_subclass(init_x.__class__)
            if isinstance(init_x, bt.Tensor): return r.special_from(init_x)
            return r
        return convert(ret)

class RNN(nn.Module):
    '''
    Recurrent Neural Network structue. 
    
    Args:
        dimension (int): The dimension of the images. 
        layers (int): The number recurrent layers. 
        conv_num (int): The number of continuous convolutions in each layer. Defaults to 2. 
        main_channels (int): The number of channels for the main sequence. Defaults to 2.
        hidden_channels (int): The number of channels for the hidden features. Defaults to 0, indicating no coupled network. 
        mid_channels (int): The number of channels for the hidden layers. Defaults to 10.
        kernel_size (int): The size of the convolution kernels. Defaults to 3. 
        padding (int or str): Indicate the type of padding used. Defaults to 'SAME' indicating a same output shape as the input. 
        conv_block (str): A string with possible values in ('conv', 'dense', 'residual'), indicating which kind of block for the layers: normal convolution layers, DenseBlock or ResidualBlock. 
        res_type (function): The combining type for the residual connections. It should be torch.add in most occasions. 
        activation_function (class): The activation function used after the convolution layers. Defaults to nn.ReLU. 
        final_activation (class): The activation function after the final convolution. Defaults to self.activation_function. 
        active_args (dict): The arguments for the activation function. Defaults to {}. 
        initializer (str): A string indicating the initialing strategy. Possible values are normal(0, 0.1) or uniform(-0.1, 0.1) or constant(0) (all parameters can be changed)
    '''
    
    class Softmax(nn.Module):
        def forward(self, x): return nn.functional.softmax(x, 1)
    
    def __init__(self, dimension = 2, layers = 10, conv_num = 2, kernel_size = 3, padding = 1, frequency = 50, linear_layer = False,
        main_channels = 2, hidden_channels = 0, mid_channels = None, conv_block = 'conv', res_type = bt.add, 
        activation_function = nn.ReLU, final_activation = ..., active_args = {}, initializer = "normal(0, 0.1)", regularization = None):
        super().__init__()
        
        if hidden_channels == True: hidden_channels = 64
        if mid_channels is None: mid_channels = main_channels
        
        params = dict(dimension=dimension, mid_channels=mid_channels, linear_layer=linear_layer, conv_num=conv_num, kernel_size=kernel_size, padding=padding, 
            activation_function=activation_function, final_activation=final_activation, active_args=active_args, conv_block=conv_block, res_type=res_type, initializer=initializer)
        avouch(layers >= 0, "RNN network should have fixed number of layers. ")
        if hidden_channels > 0: self.init_hidden = None
        self.step_layer = Convolution_Block(main_channels + hidden_channels, main_channels + hidden_channels, **params)
        
        self.n_layer = layers
        self.hidden_channels = hidden_channels
    
    def forward(self, input_x, init_hidden=None, n_output_step=None):
        """input_x: ({n_batch}, [n_feature], 'steps', n_1, n_2)"""
        self.memory_used = ByteSize(0)
        n_batch, n_feature, n_input_step, *shape = input_x.shape
        if self.hidden_channels > 0:
            self.init_hidden = bt.zeros({n_batch}, [self.hidden_channels], *shape)
            self.memory_used += self.init_hidden.byte_size()
        if n_output_step is None: n_output_step = n_input_step
        hidden = self.init_hidden
        output_x = None
        outputs = []
        for i in range(self.n_layer):
            if i < n_input_step: layer_input = input_x[:, :, i]
            else: layer_input = output_x
            output = self.step_layer(bt.cat(layer_input, hidden, []))
            self.memory_used += getattr(self.step_layer, 'memory_used', 0)
            output_x = output[:, :n_feature]
            hidden = output[:, n_feature:]
            if i >= self.n_layer - n_output_step: outputs.append(output_x)
        return bt.stack(outputs, '2'), hidden

class Models(bt.nn.Module):
    
    def __init__(self, **sub_models):
        super().__init__()
        self.device = sub_models.pop('device', bt.default_device())
        for name, model in sub_models.items():
            setattr(self, name, model.to(self.device))
        self.best_score = 0
    
    def initialize(self, weight, bias):
        def initializer(layer):
            if hasattr(layer, 'weight') and isinstance(layer.weight, torch.Tensor):
                initialize_model, initialize_params = parse(weight)
                getattr(layer.weight.data, f'{initialize_model}_')(*initialize_params)
            if hasattr(layer, 'bias') and isinstance(layer.bias, torch.Tensor):
                initialize_model, initialize_params = parse(bias)
                getattr(layer.bias.data, f'{initialize_model}_')(*initialize_params)
        for name in self.sub_networks:
            if not name or name.startswith('#'): continue
            getattr(self, name).apply(initializer)
            
    def load(self, directory, epoch='best', **kwargs):
        checkpoint_dir = Path(directory)
        if isinstance(epoch, str):
            candidates = [ckpt_file for ckpt_file in checkpoint_dir if ckpt_file.name.endswith(epoch)]
            if len(candidates) == 0: raise FileNotFoundError(f"Cannot find checkpoint with label {epoch}. ")
            if len(candidates) > 1: raise FileNotFoundError(f"Multiple checkpoints found with label {epoch}: {candidates}. ")
            ckpt_file = candidates[0]
            token = ckpt_file.name.replace('epoch', '').replace(epoch, '').strip().strip('_')
        else: ckpt_file = f"epoch{epoch}.ckpt"
        kwargs.setdefault('weights_only', True)
        self.load_state_dict(bt.load(checkpoint_dir / ckpt_file, map_location=self.device, **kwargs))
        return touch(lambda: eval(token), None)
            
    def save(self, directory, epoch, score=0):
        avouch(0 <= score <= 1, TypeError("Model score should be between [0, 1] in method 'save'. "))
        checkpoint_dir = Path(directory).mkdir()
        if score > self.best_score:
            for ckpt_file in checkpoint_dir:
                if ckpt_file.name.endswith('_best'): ckpt_file.remove()
            bt.save(self.state_dict(), checkpoint_dir / f"epoch{epoch}_best.ckpt")
            self.best_score = score
        bt.save(self.state_dict(), checkpoint_dir / f"epoch{epoch}.ckpt")

if __name__ == "__main__":
#    unet = U_Net(multi_arms="seperate(3)", block_channels=16)
#    print(unet(bt.rand(10, 3, 100, 100)).size())
#    print(*[x + ' ' + str(unet[i].size()) for x, i in unet], sep='\n')
    unet = U_Net(
        dimension=3, 
        in_channels=2, 
        out_channels=3, 
        block_channels=4, 
        with_softmax=False, 
        initializer="normal(0.0, 0.9)", 
#        conv_block='dense', 
#        conv_num=4, 
#        active_args={'inplace': True}
    )
    print(unet(bt.rand(10, 2, 50, 50, 50)).size())
    print(*[x + ' ' + str(unet[i].size()) for x, i in unet], sep='\n')



from pycamia import info_manager

__info__ = info_manager(
    project = 'PyCAMIA',
    package = 'pyoverload',
    author = 'Yuncheng Zhou',
    create = '2021-12',
    version = '2.0.0',
    contact = 'bertiezhou@163.com',
    keywords = ['overload'],
    description = "'pyoverload' overloads the functions by simply using typehints and adding decorator '@overload'. ",
    requires = ['pycamia'],
    update = '2023-07-06 20:58:10'
).check()
__version__ = '1.1.31'

from .utils import *
from .typehint import *
from .override import *




__info__ = dict(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2023-12",
    fileinfo = "File of python code operations, such as execute blocks or python code manipulation. ",
)

__all__ = """
    sorted_dict_repr
    dict_parse
    set_indent_len
    get_indent_len
    tab_to_indent
    no_indent
    with_indent
    get_num_indent
    python_lines
    python_line
    execblock
    add_lineno
""".split()

import re, io

indent_len = 4

def sorted_dict_repr(d:dict, order:list):
    """
    Representer of dictionary `d`, with key order `order`.
    
    Examples::
        >>> sorted_dict_repr({'a': 1, '0': 3}, ['0', 'a'])
        "{'0': 3, 'a': 1}"
    """
    return '{' + ', '.join([f"{repr(k)}: {repr(d[k])}" for k in order]) + '}'

def dict_parse(str_):
    return dict([(x.strip().split('=')[0].strip(), eval(x.strip().split('=')[1].strip())) for x in str(str_).split(',')])

def set_indent_len(int_):
    global indent_len
    indent_len = int_

def get_indent_len():
    global indent_len
    return indent_len

def tab_to_indent(str_):
    global indent_len
    return '\n'.join(' ' * ((len(l) - len(l.lstrip('\t'))) * indent_len) + l.lstrip('\t') for l in str_.split('\n'))

def no_indent(str_):
    min_indent = None
    lines = str(str_).split('\n')
    for l in lines:
        indent = re.search(r'^([ \t]*)[^ \t]', l)
        if indent is None: continue
        n_indent = len(indent.group(1))
        if min_indent is None or n_indent < min_indent: min_indent = n_indent
    return '\n'.join([l[min_indent:] for l in lines])

def with_indent(str_, n_indent):
    global indent_len
    indent = ' ' * (n_indent * indent_len)
    return '\n'.join([indent + l for l in no_indent(str_).split('\n')])

def get_num_indent(line: str):
    global indent_len
    num_indent = 0
    while line.startswith(' ' * indent_len) or line.startswith('\t'):
        if line.startswith('\t'): line = line[1:]
        else: line = line[indent_len:]
        num_indent += 1
    return num_indent

def python_lines(file, n_line=None):
    qts = '\'"'
    nesters = r"()[]{}"
    find_left = {r:l for l, r in zip(nesters[::2], nesters[1::2])}
    find_right = {l:r for l, r in zip(nesters[::2], nesters[1::2])}
    depth = {}
    lines = []
    cur_line_block = []
    while True:
        if isinstance(file, str): 
            line, *rest = file.split('\n', 1)
            if len(rest) == 0 and not line.strip(): break
            file = '\n'.join(rest)
        elif isinstance(file, io.TextIOWrapper):
            line = file.readline()
            if not line: break
            line = line.rstrip('\n')
        else: raise TypeError(f"'python_lines' only takes a file stream or a string of text.")
        skip_char = False
        for i, c in enumerate(line):
            if skip_char: skip_char = False; continue
            in_str = depth.get('"', 0) + depth.get("'", 0) + depth.get('"""', 0) + depth.get("'''", 0) > 0
            if c == '#' and not in_str: break
            if c == '\\' and not in_str:
                if depth.get(c, 0) > 0: raise SyntaxError("unexpected character after line continuation character")
                depth[c] = 1
            elif c == '\\': skip_char = True; continue
            elif c in qts:
                if depth.get(qts[1-qts.index(c)], 0) + depth.get(qts[1-qts.index(c)]*3, 0) == 0:
                    if line[i:i+3] == c*3: depth[c*3] = 1 - depth.get(c*3, 0)
                    else: depth[c] = 1 - depth.get(c, 0)
            elif c in find_right:
                depth[c] = depth.get(c, 0) + 1
            elif c in find_left:
                c_left = find_left[c]
                if depth.get(c_left, 0) <= 0: raise SyntaxError(f"unmatched {c!r}")
                depth[c_left] -= 1
        cur_line_block.append(line)
        if sum(depth.values()) == 0:
            lines.append('\n'.join(cur_line_block))
            cur_line_block = []
            if n_line is not None and len(lines) >= n_line: break
        elif depth.get('\\', 0) > 0: depth['\\'] = 0
    if sum(depth.values()) > 0:
        c = [k for k, n in depth.items() if n > 0][0]
        raise SyntaxError(f"{c!r} was never closed")
    if cur_line_block: lines.append('\n'.join(cur_line_block))
    return lines

def python_line(file): return python_lines(file, n_line=1)[0]

def execblock(code):
    """
    Execute `code` with indents eliminated. 
    
    Note: Assigning local variables in functions would fail just as built-in 
        method `exec`. Use `locals()[var_name]` instead to fetch the result. 

    Examples::
        >>> class A:
        ...     def run(self, x): return x ** 2
        ...     exec('''
        ...     def additional_method(self, x):
        ...         return self.run(x)
        ...          ''')
        ...
        Traceback (most recent call last):
            [...omitted...]
        IndentationError: unexpected indent
        >>> class A:
        ...     def run(self, x): return x ** 2
        ...     execblock('''
        ...     def additional_method(self, x):
        ...         return self.run(x) + self.run(x+1)
        ...          ''')
        ...
        >>> A().additional_method(3)
        25
    """
    from .environment import get_environ_vars
    code = no_indent(code)
    vars = get_environ_vars()
    loc_vars = vars.locals
    try: exec(code, vars.globals, loc_vars)
    except Exception as e:
        raise NameError(f"Error ({e}) in block execution (make sure the variable is defined in either the most global or the very local scopes): \n{code}")
    vars.update(loc_vars)
    
def add_lineno(code):
    str_i = lambda i: f"[{i:2d}]" if i < 100 else (f"{i:3d}]" if i < 1000 else f"{i:4d}")
    return '\n'.join(f"{str_i(i+1)}: {l}" for i, l in enumerate(code.split('\n')))




from .manager import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File to handle the exceptions.",
    help = "Handle the exceptions. `touch` swallow the error, `crashed` to see if there is an erro, `avouch` to assert with text, `Error` to create new error types. "
)

__all__ = """
    touch
    crashed
    avouch
    Error
""".split()

from .environment import get_args_expression, get_environ_vars, update_locals_by_environ
from .functions import const_function
from .strop import is_snakename, tokenize

def touch(target, default=None, args=[], kwargs={}, error_type: (tuple, Exception)=Exception, print_error_message=False): 
    """
    Touch a function or an expression `v`, see if it causes exception in `error_type`. 
    If not, output the result, otherwise, output the `default` value (or return from the `default` function). 
    
    Args:
        target (callable or str): The expression (or function) to touch/run. 
        *args, **kwargs (variable arguments): The inputs of function target. 
        default (callable or value object): As the function, it takes inputs (error, *args, **kwargs) to cope with exceptions. 
            Use `default = lambda *a, **k: a[0]` to return the exception object.
        error_type (tuple or Exception): The exceptions to be caught. All exceptions by default. 
    
    Examples::
        >>> a = 0
        >>> touch(lambda: 1/a, 'failed')
        failed
        >>> touch(lambda x: 1/x, args=(a,), default = lambda e, x: f'failed for {x}')
        failed for 0
    """
    if not callable(default): default = const_function(default)
    if print_error_message: dfunc = default; default = lambda *a, **k: (print(e), dfunc(e, *a, **k))[1]
    
    if isinstance(target, str):
        update_locals_by_environ()
        try:
            res = eval(target)
            if not callable(res): return res
        except error_type as e: return default(e, *args, **kwargs)
    elif callable(target):
        try:
            return target(*args, **kwargs)
        except error_type as e: return default(e, *args, **kwargs)
    else: raise TypeError(f"The input of touch should be either a string or a callable, not {target}. ")

def avouch(v: (bool, callable), excp: (str, Exception)=""):
    """
    Assert with text. 
    
    Args:
        v (bool): the expression to be validated.
        excp (str, Exception, optional): the assertion message (AssertionError), user-designed Exception when the test fails. Defaults to the asserted expression of `v`.
        
    Examples::
        >>> a = 0
        >>> avouch(a == 1, "Variable 'a' should be 0. ")
        Traceback (most recent call last):
        ...
        AssertionError: Variable 'a' should be 0.
        >>> avouch(a == 1)
        Traceback (most recent call last):
        ...
        AssertionError: Failure in assertion 'a == 1'
        >>> # The above line of code may lead to <unreachable arg expression> in native Python IDLE.
    """
    if not v:
        if not excp:
            expr = tokenize(get_args_expression('avouch'), sep=',')[0].strip()
            if (expr is not None):
                excp = f"Failure in assertion '{expr}'"
        if isinstance(excp, str):
            excp = ' '.join(excp.split())
            raise AssertionError(excp)
        else: raise excp # Inside 'avouch': a forward of previous Exception. 

def crashed(func):
    """
    Validate whether a function `func` would crash. 
    """
    try:
        func()
    except:
        return True
    return False

def Error(name: str):
    """
    Create a temporary error by text. 

    Args:
        name (str): the name of the error; It is used to identify the error type. 

    Examples::
        >>> try:
        >>>     raise Error("TEST")()
        >>> except Error("TEST"):
        >>>     print('caught')
        ... 
        caught
        >>> raise Error('Type')("description")
        Traceback (most recent call last):
        File "<stdin>", line 1, in <module>
        pycamia.exception.TESTError: description
    """
    if not is_snakename(name):
        raise TypeError(f"Invalid name '{name}' for an error: it should be alphabets/digits/underlines without spaces (as long as all other symbols). ")
    v = get_environ_vars().globals
    error_name = f"{name}Error"
    if error_name in v: return v[error_name]
    exec(f"class {error_name}(Exception): pass")
    v[error_name] = eval(error_name)
    return eval(error_name)




try: from pycamia import info_manager
except ImportError: print("Warning: pycamia not loaded. ")

__info__ = info_manager(
    project = "PyCAMIA",
    package = "pyoverload",
    author = "Yuncheng Zhou", 
    create = "2023-09",
    fileinfo = "overload functions",
    requires = ""
)

__all__ = """
    overload
    OverloadError
""".split()

from functools import wraps
# from .utils import decorator, get_environ_vars
from .typehint import typehint
from .typehint import TypeHintError, HintTypeError, get_virtual_declaration
from .typings import *

class OverloadError(Exception): ...

def decorator(wrapper_func):
    if not callable(wrapper_func): raise TypeError(f"@decorator wrapping a non-wrapper: {wrapper_func}")
    def wrapper(*args, **kwargs):
        if len(args) == 1 and callable(args[0]) or \
            len(args) == 2 and callable(args[1]):
            func = args[-1]
            raw_func = func.__func__ if isinstance(func, method) else func
            func_name = f"{raw_func.__name__}:{wrapper_func.__name__}"
            outer_func = wraps(raw_func)(wrapper_func(*args))
            outer_func.__name__ = func_name
            outer_func.__doc__ = raw_func.__doc__
            if isinstance(func, staticmethod): trans = staticmethod
            elif isinstance(func, classmethod): trans = classmethod
            else: trans = lambda x: x
            return trans(outer_func)
        return decorator(wrapper_func(*args, **kwargs))
    return wraps(wrapper_func)(wrapper)

def wrapped(f):
    while hasattr(f, '__wrapped__'):
        f = f.__wrapped__
    return f

class Overloader:

    def __init__(self):
        """@overload decorator for overloading a function with multiple usages.
        The usages are given by annotations after colons.
        
        @overload find the first suitable function to run according to the order of definition.

        Examples::
            >>> @overload
            ... def func(self, pos1, pos2: str="def_val1", / , arg3: (int, float)=3, arg4=[], *args: int, kwarg5: str='extra', **kwargs) -> str:
            ...     ... # functional content
            ...
            >>> @overload
            ... def func(self, pos1, pos2: int=1, / , arg3: (int, float)=3, arg4=[], *args: int, kwarg5: str='extra', **kwargs) -> str:
            ...     ... # functional content
            ...
            >>> fun(self, 10, 20) # call the second function
        """
        self.prev_var_name = None
        self.funcset_collected = {}

    @decorator
    def __register__(self, func):
        var_name = (func.__func__ if isinstance(func, method) else func).__qualname__
        if var_name.endswith('._'):
            if self.prev_var_name is None:
                raise NameError("Cannot use '_' as a function name of the first @overload.")
            var_name = self.prev_var_name
            var_name_candidates = [var_name]
        else:
            var_name_candidates = [var_name]
            parts__ = var_name.split('__')
            if len(parts__) >= 2 and parts__[-1]: var_name_candidates.append('__'.join(parts__[:-1]))
            if len(parts__) >= 3 and parts__[-2] and not parts__[-1]: var_name_candidates.extend(['__'.join(parts__[:-2] + ['']), '__'.join(parts__[:-2])])
            parts_ = var_name.split('_')
            if len(parts_) >= 3 and parts_[-2] and not parts_[-1]: var_name_candidates.extend(['_'.join(parts_[:-2] + ['']), '_'.join(parts_[:-2])])
        
        if not func.__name__.endswith(':typehint'): func = typehint(func, check_return=True)
        func_index = None
        for v in var_name_candidates:
            if v in self.funcset_collected:
                if v != var_name: func_index = len(self.funcset_collected[v])
                self.funcset_collected[v].append(func)
                var_name = v
                break
        else:
            self.funcset_collected[var_name] = [func]
        @wraps(func)
        def wrapper(*args, **kwargs):
            return self.__call__(*args, __cur__func__ = (var_name, func_index), **kwargs)
        return wrapper

    def __call__(self, *args, __cur__func__=None, **kwargs):
        if len(args) == 1 and not kwargs and not __cur__func__ and isinstance(args[0], callable):
            return self.__register__(args[0])
        if __cur__func__ is None: raise TypeError("Invalid @overload sub-function without keyword argument '__cur__func__'")
        if __cur__func__[1] is not None:
            # direct call
            func = self.funcset_collected[__cur__func__[0]][__cur__func__[1]]
            try: return func(*args, **kwargs)
            except (HintTypeError, TypeHintError) as e:
                args_str = ', '.join([repr(x) for x in args] + ['='.join((str(x[0]), repr(x[1]))) for x in kwargs.items()])
                raise OverloadError(f"Failed in direct call of {wrapped(func).__name__}() with arguments {args_str}. The usage is:\n" + get_virtual_declaration(wrapped(func)) + f": [{e.__class__.__name__}] {str(e)}")
        func_list = self.funcset_collected[__cur__func__[0]]
        
        failed_func_list = []
        for i, func in enumerate(func_list):
            # return func(*args, **kwargs)
            try: return func(*args, **kwargs)
            except (TypeError, HintTypeError, TypeHintError) as e:
                if e.__class__.__name__.startswith('Hint'): error_name = e.__class__.__name__[4:]
                else: error_name = e.__class__.__name__
                failed_func_list.append((func, f": [{error_name}] {str(e)}"))
        args_str = ', '.join([repr(x) for x in args] + ['='.join((str(x[0]), repr(x[1]))) for x in kwargs.items()])
        raise OverloadError(f"No {__cur__func__[0]}() matches arguments {args_str}. All available usages are:\n" + '\n'.join([get_virtual_declaration(wrapped(f)) + e for f, e in failed_func_list]))

overload = Overloader()

# @decorator
# def overload(func):
#     # Usage @overload
#     func_name = func.__name__.split('[')[0] # function name without decorator sign
#     base_func_name = [func_name[:len(func_name)-len(tag)] for tag in ('_0', "__default", '') if func_name.endswith(tag)][0] # raw function name without default tag
#     local_vars = get_environ_vars(pivot='overload', offset=2).locals
#     overload_index = f"__{base_func_name}_overload_index__"
#     overload_function = f"__{base_func_name}_overload_function__"
#     if overload_index in local_vars:
#         local_vars[overload_index] = local_vars[overload_index] + 1
#         new_name = f"__{base_func_name}_overload{local_vars[overload_index]}__"
#         local_vars[new_name] = local_vars[overload_function](func)
#     else:
#         local_vars[overload_index] = 0
#         local_vars[overload_function] = overload_wrapper(func)
#     exec(f"def {base_func_name}(*args, **kwargs): return {overload_function}(*args, **kwargs)", local_vars)
#     return local_vars[base_func_name]

# class overload_kernel:

#     def __init__(self, argfunc, first_as_default_backup=False):
#         self.main_func = argfunc
#         self.func_list = [argfunc]
#         if isinstance(argfunc, method): argfunc = argfunc.__func__
#         func_name = argfunc.__name__.split('[')[0]
#         self.base_func_name = [func_name[:len(func_name)-len(tag)] for tag in ('_0', "__default", '') if func_name.endswith(tag)][0]
#         if func_name.endswith('_0') or func_name.endswith("__default"): self.default = 0
#         else: self.default = None
#         self.first_as_default_backup = first_as_default_backup

#     def __call__(self, *args, **kwargs):
#         # Usage 1: call to register a new function
#         if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], functional):
#             argfunc = args[0]
#             func_name = argfunc.__name__.split('[')[0]
#             if func_name == '_' or func_name.startswith(self.base_func_name):
#                 if func_name.endswith('_0') or func_name.endswith("__default"):
#                     raise OverloadError("Only one default function is acceptable, please pay attention to function names ending with '_0' or '__default'. ")
#                     self.default = len(self.func_list)
#                 self.func_list.append(argfunc)
#         # Usage 2: call to find the proper overload and run
#         else:
#             declaration_list = []
#             if self.first_as_default_backup and self.default is None: self.default = 0
#             for i, func in list(enumerate(self.func_list)) + ([(-1, self.func_list[self.default])] if self.default is not None else []):
#                 if i == self.default: continue
#                 try:
#                     typehint(func, check_annot_only=True)(*args, **kwargs)
#                     return typehint(func, check_return=True)(*args, **kwargs)
#                 except (TypeError, TypeHintError) as e:
#                     error_str = f": [{e.__class__.__name__}] {str(e)}"
#                     if i == -1: declaration_list = [get_virtual_declaration(func) + error_str] + declaration_list
#                     else: declaration_list.append(get_virtual_declaration(func) + error_str)
#             args_str = ', '.join([repr(x) for x in args] + ['='.join((str(x[0]), repr(x[1]))) for x in kwargs.items()])
#             raise OverloadError(f"No {self.base_func_name}() matches arguments {args_str}. All available usages are:\n" + '\n'.join(declaration_list))

# @decorator
# def overload_wrapper(func):
#     # Usage: @overload_wrapper
#     if not isinstance(func, functional): raise TypeError("Wrong usage of @overload_wrapper: need to be used as a decorator. ")

#     overloaded_func = overload_kernel(func)
#     @wraps(func)
#     def final_wrapper(*args, **kwargs): return overloaded_func(*args, **kwargs)
#     return final_wrapper
